from ai_app.utils import PydanticForbidExtra


class AppConfig(PydanticForbidExtra):
    show: bool = False


class JiraServiceDesk(AppConfig):
    jira_requests_spreadsheet_id: str | None = None
    n_closest_requests: int = 5
    service_desks: list[int] | None = None


class JiraManualTests(AppConfig): ...


class SoftwareRequirementsSpecification(AppConfig): ...


class VoiceAssistant(AppConfig): ...


class GovernanceAgent(AppConfig): ...


class MergeRequestReview(AppConfig): ...


class SqlAgent(AppConfig): ...


class PresentationReview(AppConfig): ...


class Apps(PydanticForbidExtra):
    jira_service_desk: JiraServiceDesk = JiraServiceDesk()
    jira_manual_tests: JiraManualTests = JiraManualTests()
    software_requirements_specification: SoftwareRequirementsSpecification = (
        SoftwareRequirementsSpecification()
    )
    voice_assistant: VoiceAssistant = VoiceAssistant()
    merge_request_review: MergeRequestReview = MergeRequestReview()
    sql_agent: SqlAgent = SqlAgent()
    presentation_review: PresentationReview = PresentationReview()
    governance_agent: GovernanceAgent = GovernanceAgent()



from typing import override

import bir_mcp.config
import bir_mcp.sql
import langchain.tools

from ai_app.tools.base import BaseToolkit


class SqlToolkit(BaseToolkit):
    def __init__(
        self,
        sql_context: bir_mcp.config.SqlContext,
        max_tool_output_length: int | None = None,
    ):
        self.mcp = bir_mcp.sql.SQL(sql_context=sql_context)
        self.max_tool_output_length = max_tool_output_length

    @override
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        tools = self.mcp.get_langchain_tools(
            max_output_length=self.max_tool_output_length,
            with_prefix=False,
        )
        return tools



import logging
from typing import override

import gradio as gr
import langchain_core.language_models
import langchain_core.messages
import langsmith

from ai_app.apps.jira_service_desk.backend import Bot
from ai_app.config import get_config
from ai_app.core import BaseApp, State
from ai_app.frontend import build_model_choice_dropdown


class App(BaseApp):
    name = "Jira service desk"
    requires_auth = False

    def __init__(self):
        self.bot = Bot()

    async def astream(
        self,
        model: langchain_core.language_models.BaseChatModel,
        message: langchain_core.messages.BaseMessage,
        frontend_messages: list[langchain_core.messages.BaseMessage],
        state: State,
    ):
        messages = state.messages + [message]
        async for response in self.bot.astream(model, messages, state=state):
            yield response

    def like(self, state: State, like_data: gr.LikeData):
        if not state.messages:
            logging.warning("No messages in state to like.")
            return

        if not isinstance(like_data.liked, bool):
            # logging.info(f"Received non-bool like data: {like_data.liked}")
            # like_data.liked can be an empty string when user cancels the like.
            return

        client = langsmith.Client()
        # Index starts from user messages and alternates with AI messages.
        ai_index = like_data.index // 2
        ai_messages = [
            m for m in state.messages if isinstance(m, langchain_core.messages.AIMessage)
        ]
        ai_message = ai_messages[max(ai_index, len(ai_messages) - 1)]
        run_id = ai_message.id.removeprefix("run--").split()[0]
        client.create_feedback(
            run_id=run_id,
            key="user_like",
            score=int(like_data.liked),
        )

    @override
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks() as app:
            model_choice = model_choice or build_model_choice_dropdown()
            chat_interface = self.build_gradio_chat_interface(
                model_choice=model_choice,
                description="The chatbot will help the user find appropriate Jira request type.",
                examples=(
                    None
                    if get_config().is_prod
                    else [
                        ["I want to change my password"],
                        ["I need access to a database"],
                        ["I received a phishing email"],
                        ["A system requires additional processing resources"],
                    ]
                ),
            )
            # Registering the event handler enables the flagging buttons.
            chat_interface.chatbot.like(self.like, self.state)

        return app



import fastapi.middleware.cors
import fastapi.responses
import gradio as gr
import logfire

from ai_app.auth import mount_auth_for_gradio_app
from ai_app.config import get_config, get_favicon_path

from .gradio import build_page
from .middleware import add_middleware


def build_app() -> fastapi.FastAPI:
    app = fastapi.FastAPI()
    config = get_config()
    config.setup_logging()
    oauth = config.build_oauth()
    no_auth_params = config.get_gradio_app_kwargs(auth=False)
    if not oauth:
        gradio_app = build_page()
        app = gr.mount_gradio_app(app, gradio_app, config.gradio_app_route, **no_auth_params)

        @app.get("/")
        def root():
            return fastapi.responses.RedirectResponse(url=config.gradio_app_route)

    else:
        gradio_app = build_page(assume_auth=False)
        gradio_app_auth = build_page(assume_auth=True)
        auth_params = config.get_gradio_app_kwargs(auth=True)
        app = mount_auth_for_gradio_app(
            app=app,
            oauth_client=oauth.starlette_client,
            gradio_app=gradio_app_auth,
            login_gradio_app=gradio_app,
            login_gradio_app_route=config.gradio_login_app_route,
            gradio_app_route=config.gradio_app_route,
            auth_redirect_route=config.auth_route,
            end_session_endpoint=oauth.openid_metadata["end_session_endpoint"],
            max_id_token_size=config.max_id_token_size,
            login_gradio_app_kwargs=no_auth_params,
            gradio_app_kwargs=auth_params,
            userinfo_fields=config.user_info_fields,
        )

    @app.get("/favicon.ico", include_in_schema=False)
    async def favicon():
        return fastapi.responses.FileResponse(get_favicon_path())

    add_middleware(app)
    logfire.instrument_fastapi(app, excluded_urls=[".*/gradio_api/heartbeat.*"])
    return app



import enum
import functools
import os
import pathlib

import alembic.config
import atlassian
import bir_mcp.config
import httpx
import langchain.chat_models
import logfire
import pydantic
import sqlalchemy as sa
import yaml

import ai_app
from ai_app.ai_utils import set_langsmith_environment_variables
from ai_app.auth import OAuth
from ai_app.config.ai import ChatModelParameters
from ai_app.config.apps import Apps
from ai_app.config.external import (
    ConfluenceParameters,
    GitLabParameters,
    JiraParameters,
    SonarQubeParameters,
)
from ai_app.config.middleware import CorsConfig, EndpointRateLimitConfig, RateLimitConfig
from ai_app.config.secrets import Secrets
from ai_app.config.sql import Connection, ConnectionContext
from ai_app.external.consul import ConsulKeyValue
from ai_app.external.google import Spreadsheets, build_spreadsheets
from ai_app.external.providers import ModelProvider
from ai_app.frontend import filter_common_model_names
from ai_app.pii import AnonCodec, EntityTypeCodec
from ai_app.tools import ConfluenceToolkit, GitLabToolkit, JiraToolkit, SonarQubeToolkit, SqlToolkit
from ai_app.utils import (
    PydanticForbidExtra,
    get_module_path,
    join_url,
    refreshing_cache,
    setup_logging,
)


def get_project_src_path() -> pathlib.Path:
    return get_module_path(ai_app).parent


def to_absolute_path(path: pathlib.Path | str) -> pathlib.Path:
    path = pathlib.Path(path)
    if not path.is_absolute():
        path = get_project_src_path() / path

    path = path.resolve()
    return path


class Stage(enum.StrEnum):
    local = enum.auto()
    dev = enum.auto()
    preprod = enum.auto()
    prod = enum.auto()


class ObservabilityPlatform(enum.StrEnum):
    langsmith = enum.auto()
    langfuse = enum.auto()
    opik = enum.auto()


def get_favicon_path():
    return to_absolute_path("static/favicon_1.png")


class Config(PydanticForbidExtra):
    stage: Stage
    secrets: Secrets
    jira_parameters: JiraParameters
    confluence_parameters: ConfluenceParameters
    gitlab_parameters: GitLabParameters
    sonarqube_parameters: SonarQubeParameters
    observability_platform: ObservabilityPlatform | None = None
    enable_logfire: bool = True
    instrument_sqlalchemy: bool = False
    project: str = "ai_app"
    model_choices: list[str] = filter_common_model_names(
        max_cost=5, supports_system_messages=True, providers=[ModelProvider.openai]
    )
    azure_endpoint: str = "https://itopsai.openai.azure.com/"
    azure_api_version: str = "2024-12-01-preview"
    sanitation_deny_list: list[str] = []
    sanitation_replacing_format: str = "{{{entity_type}_{index}}}"  # Angular brackets get mixed up with HTML tags in Gradio messages.
    apps: Apps
    port: int = 7860
    gradio_login_app_route: str = "/gradio-login"
    gradio_app_route: str = "/gradio"
    auth_route: str = "/auth"
    public_url: str | None = None
    oauth_provider_url: str | None = None
    max_id_token_size: int = 0
    cors_config: CorsConfig = CorsConfig()
    file_size_mb_limit_no_auth: int = 5
    file_size_mb_limit_auth: int = 50
    security_headers: dict[str, str] = {
        "X-Frame-Options": "SAMEORIGIN",
        "Content-Security-Policy": "frame-ancestors 'self';",
        "X-Content-Type-Options": "nosniff",
        "X-XSS-Protection": "1; mode=block",
        "Referrer-Policy": "strict-origin-when-cross-origin",
    }
    user_info_fields: list[str] = ["employeeID", "preferred_username", "name", "email"]
    sql_connections: dict[str, Connection] = {}
    sql_connection_contexts: dict[str, ConnectionContext] = {}
    # When switching between chats, the gr.State gets shared between them, which may lead to context
    # leaking. To fix it, need to store a state for each thread_id.
    common_chat_interface_parameters: dict = dict(save_history=False)
    rate_limit_config: RateLimitConfig = RateLimitConfig()
    timezone: str = "Asia/Baku"
    max_tool_output_length: int | None = None
    ca_file: str | None = None

    @property
    def project_with_stage(self) -> str:
        return f"{self.project}_{self.stage}"

    @property
    def is_prod(self) -> bool:
        return self.stage == Stage.prod

    @property
    def is_local(self) -> bool:
        return self.stage == Stage.local

    @property
    def redirect_uri(self) -> str:
        url = self.public_url if self.is_prod else "http://localhost"
        url = join_url(url, self.auth_route)
        return url

    @pydantic.model_validator(mode="after")
    def _validate(cls, self):
        for connection_context in self.sql_connection_contexts.values():
            if connection_context.connection_name not in self.sql_connections:
                raise ValueError(
                    f"Connection '{connection_context.connection_name}' not found in 'sql_connections'"
                )

        return self

    def model_post_init(self, _context):
        if self.is_local:
            self.cors_config.allow_origin_regex = "http://localhost:.*"
            self.cors_config.allow_credentials = True

        if self.public_url:
            self.cors_config.allow_origins.append(self.public_url)

        if self.ca_file:
            bir_mcp.utils.set_ssl_cert_file_from_cadata(self.ca_file)

    def setup_logging(self):
        """
        Calling this function twice may lead to an error.
        Httpx and fastapi packages need to be instrumented per client and app respectively:
            logfire.instrument_fastapi(app)
        """
        match self.observability_platform:
            case ObservabilityPlatform.langsmith:
                set_langsmith_environment_variables(
                    langsmith_api_key=self.secrets.langsmith_api_key.get_secret_value(),
                    langsmith_project=self.project_with_stage,
                )

        if self.enable_logfire:
            logfire.configure(
                environment=str(self.stage),
                token=self.secrets.logfire_api_token.get_secret_value(),
                scrubbing=False if self.is_local else None,
                send_to_logfire=True,
            )
            logfire.instrument_system_metrics()
            logfire.instrument_pydantic(record="failure")
            logfire.instrument_httpx(
                capture_request_body=self.is_local,
                capture_response_body=self.is_local,
            )
            if self.instrument_sqlalchemy:  # if self.is_local:
                # Logs every SQL query, which is overkill for most cases.
                logfire.instrument_sqlalchemy()

        setup_logging(handlers=[logfire.LogfireLoggingHandler()])

    @staticmethod
    def from_consul():
        consul_key_value = ConsulKeyValue.from_env_variables()
        value = consul_key_value.load()
        config = yaml.safe_load(value)
        config = Config(**config)
        return config

    def get_jira_kwargs(self) -> dict:
        kwargs = self.jira_parameters.model_dump()
        if self.secrets.jira_api_token:
            kwargs["token"] = self.secrets.jira_api_token.get_secret_value()

        return kwargs

    def get_jira_toolkit_kwargs(self) -> dict:
        kwargs = self.get_jira_kwargs()
        kwargs |= self.model_dump(include={"max_tool_output_length"})
        return kwargs

    def get_confluence_toolkit_kwargs(self) -> dict:
        kwargs = self.confluence_parameters.model_dump()
        kwargs |= self.model_dump(include={"max_tool_output_length"})
        if self.secrets.confluence_api_token:
            kwargs["token"] = self.secrets.confluence_api_token.get_secret_value()

        return kwargs

    def get_gitlab_kwargs(self) -> dict:
        kwargs = self.gitlab_parameters.model_dump()
        kwargs |= self.model_dump(include={"timezone", "max_tool_output_length"})
        if self.secrets.gitlab_api_key:
            kwargs["private_token"] = self.secrets.gitlab_api_key.get_secret_value()

        return kwargs

    def get_sonarqube_kwargs(self) -> dict:
        kwargs = self.sonarqube_parameters.model_dump()
        kwargs |= self.model_dump(include=["timezone", "max_tool_output_length"])
        kwargs["gitlab_url"] = self.gitlab_parameters.url
        if self.secrets.sonarqube_api_key:
            kwargs["token"] = self.secrets.sonarqube_api_key.get_secret_value()

        return kwargs

    def build_oauth(self) -> OAuth | None:
        if not self.oauth_provider_url:
            return

        # Verify provider reachability (timeout/error -> skip OAuth setup).
        try:
            response = httpx.get(self.oauth_provider_url, timeout=5.0)
            response.raise_for_status()
        except (httpx.RequestError, httpx.HTTPStatusError):
            return

        oauth = OAuth(
            provider_base_url=self.oauth_provider_url,
            client_id=self.secrets.oauth_client_id.get_secret_value(),
            client_secret=self.secrets.oauth_client_secret.get_secret_value(),
            redirect_uri=self.redirect_uri,
        )
        return oauth

    def get_gradio_app_kwargs(self, auth: bool = False) -> dict:
        max_file_size_mb = self.file_size_mb_limit_auth if auth else self.file_size_mb_limit_no_auth
        kwargs = dict(
            max_file_size=f"{max_file_size_mb}mb",
            show_error=not self.is_prod,
            enable_monitoring=not self.is_prod,
            allowed_paths=[get_favicon_path()],
        )
        return kwargs

    def build_entity_codec(self) -> EntityTypeCodec:
        codec = EntityTypeCodec(replacing_format=self.sanitation_replacing_format)
        return codec

    def get_provider_api_key(self, provider: ModelProvider) -> pydantic.SecretStr | None:
        provider_api_keys = self.secrets.provider_api_keys
        api_key = provider_api_keys.get(provider)
        return api_key

    def build_chat_model_parameters(self, model: str):
        parameters = ChatModelParameters.from_model_name(model)
        parameters.api_key = self.get_provider_api_key(parameters.model_provider)
        return parameters


@refreshing_cache()
def get_config() -> Config:
    try:
        config = Config.from_consul()
    except KeyError:
        config_path = os.getenv("CONFIG_PATH")
        with open(config_path) as f:
            config = yaml.safe_load(f)

        config = Config(**config)

    return config


def get_jira_kwargs() -> dict:
    return get_config().get_jira_kwargs()


def get_gitlab_kwargs() -> dict:
    return get_config().get_gitlab_kwargs()


@functools.cache
def get_chat_model(model_name: str):
    config = get_config()
    if model_name.startswith("azure"):
        model_name = model_name.removeprefix("azure-")
        provider = ModelProvider.azure_openai
        model = langchain.chat_models.init_chat_model(
            model_name,
            model_provider=provider,
            azure_endpoint=config.azure_endpoint,
            api_version=config.azure_api_version,
            azure_deployment=model_name,
            api_key=config.get_provider_api_key(provider),
        )
        return model

    return config.build_chat_model_parameters(model_name).build_model()


@functools.cache
def get_engine(connection_name: str) -> sa.Engine:
    return get_config().sql_connections[connection_name].build_engine()


@functools.cache
def get_ai_postgres_engine() -> sa.Engine:
    return get_engine("postgres_ai")


@functools.cache
def get_ai_postgres_rag_engine() -> sa.Engine:
    return get_engine("postgres_ai_rag")


@functools.cache
def get_jira():
    return atlassian.Jira(**get_jira_kwargs())


@functools.cache
def get_service_desk():
    return atlassian.ServiceDesk(**get_jira_kwargs())


@functools.cache
def get_jira_toolkit():
    return JiraToolkit(**get_config().get_jira_toolkit_kwargs())


@functools.cache
def get_confluence_toolkit():
    return ConfluenceToolkit(**get_config().get_confluence_toolkit_kwargs())


@functools.cache
def get_gitlab_toolkit():
    return GitLabToolkit(**get_gitlab_kwargs())


@functools.cache
def get_sonarqube_toolkit():
    return SonarQubeToolkit(**get_config().get_sonarqube_kwargs())


@functools.cache
def get_anon_codec():
    return AnonCodec(deny_list=get_config().sanitation_deny_list)


@functools.cache
def get_sql_toolkit(connection_context_name: str) -> SqlToolkit:
    connection_context = get_config().sql_connection_contexts[connection_context_name]
    connection = get_config().sql_connections[connection_context.connection_name]
    sql_context = bir_mcp.config.SqlContext(
        connection=connection.model_dump(),
        schema_tables=connection_context.schema_tables,
    )
    toolkit = SqlToolkit(
        sql_context=sql_context,
        max_tool_output_length=get_config().max_tool_output_length,
    )
    return toolkit


def get_alembic_config(
    alembic_ini: str = "alembic.ini", alembic_directory: str = "alembic"
) -> alembic.config.Config:
    project_root_path = get_project_src_path().parent
    config = alembic.config.Config(project_root_path / alembic_ini)
    config.set_main_option("script_location", str(project_root_path / alembic_directory))
    return config


@functools.cache
def get_jira_service_desk_spreadsheets() -> Spreadsheets:
    spreadsheets = build_spreadsheets(
        spreadsheet_id=get_config().apps.jira_service_desk.jira_requests_spreadsheet_id,
        service_account_info=get_config().secrets.google_service_account_info.get_secret_value(),
    )
    return spreadsheets


def get_judge_model_names() -> list[str]:
    models = [
        "gpt-4.1",
        "gpt-5",
        # "gemini-2.5-pro",
        # "anthropic/claude-4-sonnet",
    ]
    return models



import enum

import atlassian
import gradio as gr
import langchain_core
import langsmith
import pydantic
import requests

from ai_app.config import get_chat_model, get_jira_kwargs
from ai_app.external.atlassian import extract_jira_issue_fields


class ProvidedInformationEvaluation(pydantic.BaseModel):
    """The evaluation of the provided information, how clear, complete and consitent it is, whether or not it enables creation of good manual tests, and why."""

    feedback: str = pydantic.Field(
        description=(
            "The feedback on the quality of the provided information, how clear, complete and consitent it is, "
            "whether or not it enables creation of good manual tests, and why."
        )
    )
    improvement_request: str = pydantic.Field(
        description=(
            "If there is a clear way to improve the quality of the provided information, or if some important information was not provided, "
            "then that information should be included in this improvement request."
        )
    )
    information_quality: int = pydantic.Field(
        description=(
            "The quality of the information provided regarding the Jira issue and what needs to be tested, "
            "on a scale from 1 to 5, where 1 is very poor and 5 is very good."
        )
    )


class JiraTestStep(pydantic.BaseModel):
    """One of the test steps in a sequence of steps comprising the actionable part of the Jira test."""

    purpose: str | None = pydantic.Field(
        default=None, description="One sentence describing the purpose of this test step."
    )
    action: str = pydantic.Field(description="The action that the human tester should perform.")
    expected_result: str = pydantic.Field(description="The expected result following the action.")


class TestType(enum.Enum):
    """The type of a Jira manual test."""

    new_functional = "A functional test that verifies that new features are working as expected."
    smoke_regress = "A preliminary test that quickly checks whether critical functionalities still work after changes."
    smart_regress = "A targeted test that selectively retests the most impacted or high-risk areas of an application."
    full_regress = (
        "A comprehensive test that verifies all existing functionalities of an application."
    )
    draft = "A test category used for cases that do not align with other predefined test types."

    def to_jira_field(self) -> str:
        field = self.name.replace("_", " ").capitalize()
        return field

    @classmethod
    def from_jira_field(cls, test_type: str):
        test_types = {t.to_jira_field(): t for t in TestType}
        test_type = test_types[test_type]
        return cls(test_type)


class JiraTest(pydantic.BaseModel):
    """A manual test related to some Jira issue, meant to verify the quality of whatever the issue represents."""

    title: str = pydantic.Field(description="The title for the test")
    short_description: str = pydantic.Field(
        description=(
            "The description for the test. "
            "If testing a bug, it should include explanation how the test prevents the bug from reoccuring."
            "If testing new features or changes, it should describe how each feature should be tested, "
            "mention possible security vulnerabilities, and assess most likely points of failure and their severity."
        )
    )
    steps: list[JiraTestStep] = pydantic.Field(
        description="The list of approximately 10 test steps that the human tester should manually execute."
    )
    type: TestType = TestType.draft

    def build_creation_fields(self, project: str, assignee_name: str) -> dict:
        steps = [
            {"Action": step.action, "Data": "", "Expected Result": step.expected_result}
            for step in self.steps
        ]
        steps = {"steps": [{"fields": step} for step in steps]}
        parameters = {
            "summary": self.title,
            "description": self.short_description,
            "project": {"key": project},
            "issuetype": {"name": "Test"},
            "assignee": {"name": assignee_name},
            "customfield_18504": steps,
            "customfield_23710": {"value": self.type.to_jira_field()},
        }
        return parameters


class JiraTestWithEvaluation(pydantic.BaseModel):
    evaluation: ProvidedInformationEvaluation
    jira_test: JiraTest


def build_prompt_template():
    prompt_template = langchain_core.prompts.ChatPromptTemplate(
        [
            (
                "system",
                "You are an expert at writing manual tests based on issues in Jira.",
            ),
            (
                "human",
                """
You will be provided with attributes for a single Jira issue, including the issue type, title, and description,
along with an additional prompt that contains specific details and instructions. Some of the information may be
partially written in Azerbaijani. Your tasks are as follows:

1. Evaluate the quality of the provided information. If any information is lacking or unclear, ask for clarification.
2. Rate the quality of the information on a scale from 1 to 5, where 1 represents very poor quality and 5 represents very good quality.
3. Based on the evaluated information, write a manual test for the provided Jira issue:
    - If the Jira issue describes a bug in code or another defect, create a test designed to detect that bug.
    - If the Jira issue describes new features in code, an application, or a user interface, create a test that 
verifies the new features function as expected and that the changes do not introduce any security vulnerabilities or unexpected side effects.

In the context of manual testing, the following abbreviations may be used:
- STR: Steps to reproduce
- AR: Actual result
- ER: Expected result

The output test should include:
- A title
- A description
- A list of test steps, each consisting of a purpose, an action and an expected result for that action

Here are the attributes you will work with:
<jira_issue_type>{issuetype}</jira_issue_type>
<jira_issue_title>{summary}</jira_issue_title>
<jira_issue_description>{description}</jira_issue_description>

And here is the additional prompt from the end user:
<additional_prompt>{additional_prompt}</additional_prompt>
""",
            ),
        ]
    )
    return prompt_template


class JiraTestGenerator:
    def __init__(
        self,
        max_jira_issue_field_length: int = 10_000,
        langsmith_trace: str = "Generate Jira manual test",
    ):
        self.max_jira_issue_field_length = max_jira_issue_field_length
        self.langsmith_trace = langsmith_trace
        self.jira = atlassian.Jira(**get_jira_kwargs())
        self.prompt_template = build_prompt_template()

    def generate(
        self, model_name: str, issue_key: str, additional_prompt: str | None = None
    ) -> tuple[JiraTestWithEvaluation | None, str]:
        try:
            issue = self.jira.get_issue(issue_key, fields="issuetype, summary, description")
        except requests.HTTPError as e:
            raise gr.Error(
                f"Encountered an error when trying to fetch Jira issue '{issue_key}'."
            ) from e

        fields = extract_jira_issue_fields(issue, max_length=self.max_jira_issue_field_length)
        prompt_params = fields | dict(additional_prompt=additional_prompt)
        model = get_chat_model(model_name)
        model = model.with_structured_output(JiraTestWithEvaluation)
        with langsmith.trace(self.langsmith_trace) as run:
            prompt = self.prompt_template.invoke(prompt_params)
            jira_test_with_evaluation = model.invoke(prompt)
            run_id = run.id

        return jira_test_with_evaluation, run_id

    def create_jira_test(self, jira_test: JiraTest, project: str, assignee: str) -> dict:
        fields = jira_test.build_creation_fields(project=project, assignee_name=assignee)
        try:
            issue = self.jira.create_issue(fields=fields)
        except requests.HTTPError as e:
            raise gr.Error("Encountered an error when trying to create a Jira issue.") from e

        return issue



import langchain_core
import langsmith
import pydantic

from ai_app.ai_utils import invoke_model
from ai_app.config import get_chat_model


class Characteristic(pydantic.BaseModel):
    feedback: str = pydantic.Field(
        description="Feedback on whether the specification exhibits the characteristic, and why."
    )
    grade: int = pydantic.Field(
        description=(
            "A characteristic grade is an integer value from 1 to 5, where 1 indicates that the specification does not exhibit "
            "the characteristic at all, and 5 indicates sufficient presence of the characteristic."
        )
    )


class Characteristics(pydantic.BaseModel):
    """
    A collection of characteristics that a software requirements specification should exhibit according to the IEEE standard.
    A specification may be graded on each characteristic, resulting in a grade from 1 to 5, along with feedback explaining the grade.
    """

    correct: Characteristic = pydantic.Field(
        description="A specification is correct if every requirement stated therein is one that the software shall meet.",
    )
    unambiguous: Characteristic = pydantic.Field(
        description="A specification is unambiguous if every requirement stated therein has only one interpretation.",
    )
    complete: Characteristic = pydantic.Field(
        description=(
            "A specification is complete if it includes all significant requirements, "
            "whether relating to functionality, performance, design constraints, attributes, or external interfaces."
        ),
    )
    consistent: Characteristic = pydantic.Field(
        description=(
            "A specification is consistent if it does not disagree with some other internal document."
        ),
    )
    ranked_for_importance: Characteristic = pydantic.Field(
        description=(
            "A specification is ranked for importance if each requirement in it has an identifier to indicate either "
            "the importance or stability of that particular requirement."
        ),
    )
    verifiable: Characteristic = pydantic.Field(
        description=(
            "A specification is verifiable if for every requirement stated therein exists some efficient process "
            "with which a person or machine can check that the software product meets the requirement. "
            "In general any ambiguous requirement is not verifiable."
        ),
    )
    modifiable: Characteristic = pydantic.Field(
        description=(
            "A specification is modifiable if its structure and style are such that any changes to the requirements "
            "can be made easily, completely, and consistently while retaining the structure and style."
        ),
    )
    traceable: Characteristic = pydantic.Field(
        description=(
            "A specification is traceable if the origin of each of its requirements is clear and if it facilitates "
            "the referencing of each requirement in future development or enhancement documentation"
        ),
    )


class SoftwareRequirementsSpecificationEvaluation(pydantic.BaseModel):
    """
    A collection of desirable characteristics for a software requirements specification, along with a recommendation
    for improving the specification and an example of such improvements.
    """

    characteristics: Characteristics
    specification_improvement_recommendation: str = pydantic.Field(
        description="A recommendation on how to improve the specification to better adhere to the IEEE standard.",
    )
    specification_improvement_example: str = pydantic.Field(
        description="An example of how to improve the specification, making it follow the IEEE guidelines.",
    )


def build_prompt_template() -> langchain_core.prompts.ChatPromptTemplate:
    prompt_template = langchain_core.prompts.ChatPromptTemplate(
        [
            (
                "system",
                "You are an expert at writing and evaluating software requirements specifications according to the IEEE standard.",
            ),
            (
                "human",
                """
You will be provided with a software requirements specification for a component of a digital banking ecosystem,
which may include sections partially written in Azerbaijani. Your tasks are as follows:
1. Evaluate the specification based on the IEEE standard for software requirements specifications.
2. Assess the specification against a list of characteristics, for each writing short feedback on whether or not the
specification exhibits the characterstic, and assigning an integer value from 1 to 5, where 1 indicates a complete lack
of the characteristic, and 5 indicates a sufficient presence of the characteristic.
3. If possible, suggest improvements for the specification to enhance its characteristics and overall adherence to the IEEE standard.

Here are some guidelines that the specification should follow:
1. Requirements should strictly be about what is needed, independently of the system design, and not how the software should do it.
2. Specification should avoide subjective language, ambiguous adverbs and adjectives, superlatives and negative statements. 
Comparative phrases, non-verifiable terms or terms implying totatily should also be avoided.

Here is the software requirements specification to evaluate:
<specification>{specification}</specification>
                """,
            ),
        ]
    )
    return prompt_template


class SpecificationEvaluator:
    def __init__(
        self,
        max_prompt_param_length: int = 100_000,
        max_completion_tokens: int = 4_000,
        langsmith_trace: str = "Evaluate software requirements specification",
    ):
        self.max_prompt_param_length = max_prompt_param_length
        self.max_completion_tokens = max_completion_tokens
        self.langsmith_trace = langsmith_trace
        self.prompt_template = build_prompt_template()

    def evaluate(
        self, model_name: str, specification: str
    ) -> tuple[SoftwareRequirementsSpecificationEvaluation | None, str]:
        prompt_params = dict(specification=specification)
        model = get_chat_model(model_name)
        model = model.with_structured_output(SoftwareRequirementsSpecificationEvaluation)
        evaluation, run_id = invoke_model(
            model,
            prompt_template=self.prompt_template,
            prompt_params=prompt_params,
            trace=langsmith.trace(self.langsmith_trace),
            max_prompt_param_length=self.max_prompt_param_length,
        )
        return evaluation, run_id



from ai_app.external.gitlab import GitlabUrl
from ai_app.utils import PydanticForbidExtra


class JiraParameters(PydanticForbidExtra):
    url: str = "https://jira-support.kapitalbank.az"
    backoff_and_retry: bool = True
    retry_status_codes: list[int] = [400, 401, 413, 429, 503, 504]
    max_backoff_seconds: int = 10
    max_backoff_retries: int = 10


class ConfluenceParameters(PydanticForbidExtra):
    url: str = "https://confluence.kapitalbank.az"


class GitLabParameters(PydanticForbidExtra):
    url: GitlabUrl


class SonarQubeParameters(PydanticForbidExtra):
    url: str = "https://sonarqube.kapitalbank.az"



from typing import override

import atlassian
import bir_mcp
import langchain.tools

from ai_app.external.atlassian import get_project_overview
from ai_app.tools.base import BaseToolkit


class JiraToolkit(BaseToolkit):
    def __init__(
        self,
        token: str,
        url: str,
        api_version: int = 2,
        max_tool_output_length: int | None = None,
        **kwargs,  # The kwargs from JiraParameters, need to use them for tenacity manually instead of passing to atlassian.Jira.
    ):
        super().__init__()
        self.mcp = bir_mcp.atlassian.Jira(token=token, url=url, api_version=api_version, **kwargs)
        self.max_tool_output_length = max_tool_output_length
        self.jira = atlassian.Jira(token=token, url=url, api_version=api_version, **kwargs)

    def get_jira_project_overview(self, project_key: str) -> dict:
        """Fetches the overview of a Jira project."""
        overview = get_project_overview(self.jira, project_key)
        return overview

    @override
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        tools = [
            self.mcp.get_issue_overview,
            self.get_jira_project_overview,
        ]
        tools = [
            bir_mcp.utils.to_langchain_tool(tool, max_output_length=self.max_tool_output_length)
            for tool in tools
        ]
        return tools


class ConfluenceToolkit(BaseToolkit):
    def __init__(self, token: str, url: str, max_tool_output_length: int | None = None):
        super().__init__()
        self.mcp = bir_mcp.atlassian.Confluence(token=token, url=url)
        self.max_tool_output_length = max_tool_output_length

    @override
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        tools = self.mcp.get_langchain_tools(max_output_length=self.max_tool_output_length)
        return tools



from abc import ABC, abstractmethod

import langchain.tools


class BaseToolkit(ABC):
    @abstractmethod
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        """For compatibility with langchain_core.tools.BaseToolkit."""

    def get_named_tools(self) -> dict[str, langchain.tools.BaseTool]:
        tools = self.get_tools()
        tools = {tool.name: tool for tool in tools}
        return tools



import datetime
import json
import logging
import textwrap
import warnings

import atlassian
import langchain_core.documents
import polars as pl
import pydantic
import requests
import sqlmodel
import tqdm.auto

from ai_app.ai_utils import (
    SummarizationStateBase,
    Summarizer,
    get_pydantic_config_dict_for_openai_structured_output,
    process_nested_basic_datatypes,
    to_json_llm_input,
    try_converting_to_openai_flex_tier,
)
from ai_app.apps.jira_service_desk.data import (
    AiInputRequestFields,
    AiJiraRequestGuide,
    AiJiraRequestGuideContent,
    InputRequestField,
    JiraRequest,
    JiraRequestCommonColumns,
    ManualJiraRequestGuide,
    build_row_index,
    get_index_column_names,
    get_saved_request_type_primary_keys,
)
from ai_app.apps.jira_service_desk.tools import get_postgres_vector_store
from ai_app.config import (
    get_ai_postgres_engine,
    get_anon_codec,
    get_chat_model,
    get_jira,
    get_service_desk,
)
from ai_app.external.atlassian import (
    build_request_url,
    extract_issue_field_values,
    generate_jql_issues,
    get_jira_base_url,
    get_request_field_names,
    get_request_fields,
    get_service_desks,
    get_service_desks_request_types,
    normalize_line_breaks,
)
from ai_app.external.google import Spreadsheets
from ai_app.sql import open_session


def get_request_common_field_names() -> dict[str, str]:
    fields = {
        "resolution": "Resolution",
        "reporter": "Reporter",
        "assignee": "Assignee",
    }
    return fields


def get_irrelevant_request_fields() -> dict[str, str]:
    fields = {
        "attachment": "Attachment",
        "customfield_10300": "Approvers",
        "approvers": "Approvers",
    }
    return fields


def get_request_resolution_statuses() -> set[str]:
    statuses = {
        "Done",
        "Rejected",
        "Resolved",
        "No response",
        "Won't Do",
        "Declined",
        "Unresolved",
        "Duplicate",
    }
    return statuses


# def get_service_desk_ids_to_ignore() -> set[int]:
#     ids = {
#         23,  # Pasha Bank Georgia.
#         42,  # Pasha service desk, which I don't have access to.
#         442,  # No access, Jira project TB.
#     }
#     return ids


def update_jira_requests():
    """
    Updates {JiraRequest} and {InputRequestField} tables by collecting all the request types
    for all service desks from Jira API. Delete service desks and request types that are no longer
    present in Jira or that the API key doesn't have access to.
    """
    engine = get_ai_postgres_engine()
    jira = get_jira()
    service_desk = get_service_desk()
    service_desks = set()
    for service_desk_id, desk in get_service_desks(service_desk).items():
        try:
            jira.get_project(desk["projectKey"])
        except requests.HTTPError:
            logging.warning(
                f"No access to Jira project {desk['projectKey']}, skipping service desk."
            )
        else:
            service_desks.add(service_desk_id)

    with open_session(engine, commit_on_exit=True) as session:
        session.exec(
            sqlmodel.delete(JiraRequest).where(JiraRequest.service_desk_id.notin_(service_desks))
        )

    for service_desk_id in service_desks:
        service_desks_request_types = get_service_desks_request_types(
            service_desk=service_desk,
            service_desk_id=service_desk_id,
            verbose=True,
        )
        with open_session(engine, commit_on_exit=True) as session:
            session.exec(
                sqlmodel.delete(JiraRequest).where(
                    JiraRequest.service_desk_id == service_desk_id,
                    JiraRequest.request_type_id.notin_(service_desks_request_types),
                )
            )
            for request_type_id, request_type in tqdm.auto.tqdm(
                service_desks_request_types.items()
            ):
                index = build_row_index(
                    service_desk_id=service_desk_id,
                    request_type_id=request_type_id,
                )
                request = JiraRequest(
                    **index,
                    name=request_type["name"],
                    description=request_type.get("description"),
                )
                session.merge(request)

                try:
                    request_fields_df = get_request_fields(
                        service_desk=service_desk,
                        service_desk_id=service_desk_id,
                        request_type_id=request_type_id,
                    )
                except Exception:
                    warnings.warn(
                        f"Failed to get request fields for service desk {service_desk_id} "
                        f"and request type {request_type_id}\n"
                        f"Request URL: {build_request_url(service_desk_id, request_type_id)}"
                    )
                    raise

                if request_fields_df is not None:
                    for row in request_fields_df.iter_rows(named=True):
                        field = InputRequestField(
                            **index,
                            field_id=row["fieldId"],
                            name=row["name"],
                            description=row["description"],
                            can_be_passed_as_url_parameter=row["can_be_passed_as_url_parameter"],
                            valid_values=row["valid_value_labels"],
                        )
                        session.merge(field)


def update_manual_jira_requests(
    spreadsheets: Spreadsheets,
    sheet_name: str = "Service desk requests",
    write_back_to_gsheet: bool = False,
):
    f"""Updates {ManualJiraRequestGuide} table."""

    index_columns = get_index_column_names()
    df = spreadsheets.read_sheet(sheet_name)
    df = df.select(
        *(pl.col(c).str.to_integer() for c in index_columns + ["do_recommend"]),
        "use_case",
        "resolution",
    )
    saved_request_primary_keys = get_saved_request_type_primary_keys(get_ai_postgres_engine())
    saved_request_primary_keys = {i.primary_key_tuple for i in saved_request_primary_keys}
    with open_session(get_ai_postgres_engine()) as session:
        for row in df.iter_rows(named=True):
            guide = ManualJiraRequestGuide(**row)
            if guide.primary_key_tuple not in saved_request_primary_keys:
                warnings.warn(
                    f"Request type primary key from manual description not found among saved: "
                    f"{guide.primary_key_tuple}"
                )
            else:
                session.merge(guide)

        session.commit()

    if write_back_to_gsheet:
        ai_df = pl.read_database(sqlmodel.select(AiJiraRequestGuide), get_ai_postgres_engine())
        columns_to_drop = set(JiraRequestCommonColumns.model_fields) - set(index_columns)
        ai_df = ai_df.drop(columns_to_drop)
        ai_df = ai_df.rename({c: f"ai_{c}" for c in set(ai_df.columns) - set(index_columns)})

        jira_df = pl.read_database(sqlmodel.select(JiraRequest), get_ai_postgres_engine())
        jira_df = jira_df.drop(columns_to_drop)
        base_url = get_jira_base_url()
        url = f"{base_url}/servicedesk/customer/portal/{{}}/create/{{}}"
        jira_df = jira_df.with_columns(pl.format(url, *index_columns).alias("url"))
        jira_df = jira_df.rename(
            {c: f"jira_{c}" for c in set(jira_df.columns) - set(index_columns)}
        )

        jira_df = jira_df.join(ai_df, on=index_columns, how="left")
        df = jira_df.join(df, on=index_columns, how="left")
        df = df.with_columns(pl.col("do_recommend").fill_null(0))
        spreadsheets.write_sheet(sheet_name, df)


def build_structured_output_class(
    input_attributes: list[str],
    parent_class: type[pydantic.BaseModel] = AiJiraRequestGuideContent,
) -> type:
    InputRequestFields = pydantic.create_model(
        "InputRequestFields",
        __config__=get_pydantic_config_dict_for_openai_structured_output(),
        __doc__=textwrap.dedent(
            """
            A collection of request attributes that the user needs to provide,
            with short description, common user errors (if present) and example for each
            """
        ),
        **{i: (str, ...) for i in input_attributes},
    )

    class JiraRequestUserGuide(
        parent_class, **get_pydantic_config_dict_for_openai_structured_output()
    ):
        input_request_fields: InputRequestFields

    return JiraRequestUserGuide


def fetch_issue_comments(service_desk: atlassian.ServiceDesk, issue: dict, limit_comments: int):
    issue_comments = service_desk.get_request_comments(issue["Issue key"], limit=limit_comments)
    issue_comments = sorted(issue_comments, key=lambda c: c["created"]["iso8601"])
    issue_comments = [(c["author"]["displayName"], c["body"]) for c in issue_comments]
    participants = {
        issue["Reporter"]: "Reporter",
        issue["Assignee"]: "Assignee",
    }
    issue_comments = [
        (participants.get(author, author), normalize_line_breaks(body))
        for author, body in issue_comments
    ]
    issue_comments = [f"{author}: {body}" for author, body in issue_comments]
    return issue_comments


def generate_requests(
    service_desk_id: int,
    request_type_id: int,
    jql_filter: str | None = None,
    fields: dict[str, str] | None = None,
    lookbehind: datetime.timedelta = datetime.timedelta(days=365),
    limit_requests: int = 100,
    limit_comments: int = 50,
    verbose: bool = False,
):
    jira = get_jira()
    service_desk = get_service_desk()
    fields = dict(fields or get_request_field_names(service_desk, service_desk_id, request_type_id))
    fields |= get_request_common_field_names()
    fields = {k: v for k, v in fields.items() if k not in get_irrelevant_request_fields()}
    service_desk_name_project = get_service_desks(service_desk)[service_desk_id]["projectKey"]
    request_type_name = get_service_desks_request_types(service_desk, service_desk_id)[
        request_type_id
    ]["name"]
    request_type_name = request_type_name.replace("\t", r"\t")
    jql_filter = jql_filter or textwrap.dedent(f"""
        resolution IS NOT EMPTY
        AND resolution NOT IN ('No response')
        AND resolutiondate > {datetime.date.today() - lookbehind}
    """)
    jql = f"""
        project = {service_desk_name_project}
        AND 'customer request type' = '{request_type_name} ({service_desk_name_project})'
        AND ({jql_filter})
        ORDER BY resolutiondate DESC
    """  # noqa: E501
    for issue in generate_jql_issues(
        jira=jira, jql=jql, fields=list(fields), soft_limit=limit_requests, verbose=verbose
    ):
        issue = extract_issue_field_values(issue)
        issue = {fields.get(k, k): v for k, v in issue.items()}
        issue["Comments"] = fetch_issue_comments(service_desk, issue, limit_comments)
        yield issue


def get_request(issue_key: str) -> dict:
    jira = get_jira()
    issue = jira.get_issue(issue_key)
    return issue


def update_ai_jira_request_guides(service_desk_ids: list[int] | None = None, **kwargs):
    f"""Updates {AiJiraRequestGuide} table."""
    # TODO: convert to async, use asyncio.TaskGroup for concurrent execution.
    statement = sqlmodel.select(JiraRequest)
    if service_desk_ids:
        statement = statement.where(JiraRequest.service_desk_id.in_(service_desk_ids))

    with open_session(get_ai_postgres_engine()) as session:
        for request in tqdm.auto.tqdm(
            session.exec(statement),
            desc="Iterating over Jira request types from the JiraRequest table",
        ):
            summarize_request_type(request.service_desk_id, request.request_type_id, **kwargs)


class SummarizationState(AiJiraRequestGuideContent, SummarizationStateBase): ...


def summarize_request_type(
    service_desk_id: int, request_type_id: int, model=None, max_tokens: int = 800_000, **kwargs
) -> AiJiraRequestGuide | None:
    index = build_row_index(service_desk_id=service_desk_id, request_type_id=request_type_id)
    output_attributes = ["Resolution", "Comments"]
    meta_attributes = ["Issue key", "Reporter", "Assignee", "Approvers"]
    fields = get_request_field_names(get_service_desk(), service_desk_id, request_type_id)
    input_attributes = list(set(fields.values()) - set(output_attributes) - set(meta_attributes))
    SummarizationStateWithFields = build_structured_output_class(
        input_attributes=input_attributes,
        parent_class=SummarizationState,
    )
    if not model:
        model = get_chat_model("gpt-5-mini")
        model = try_converting_to_openai_flex_tier(model)
        # See docs for more OpenAI API parameters:
        # https://platform.openai.com/docs/api-reference/responses/create#responses_create-reasoning
        model = model.model_copy(
            update={
                "reasoning": {
                    "effort": "medium",
                    # "summary": "detailed", # Requires the OpenAI organization to be verified.
                },
            }
        )

    summarizer = Summarizer(
        model=model,
        max_tokens=max_tokens,
        context=build_summarization_context_prompt(**index, input_attributes=input_attributes),
    )
    items = generate_requests(**index, fields=fields, verbose=True, **kwargs)
    anon_codec = get_anon_codec()
    items = process_nested_basic_datatypes(items, process_string=anon_codec.anonymize)
    items = [json.dumps(i, ensure_ascii=False) for i in items]
    state = summarizer.summarize(state=SummarizationStateWithFields, items=items)
    if not state:
        return

    ai_request_guide = AiJiraRequestGuide(
        **index,
        **state.model_dump(include=list(AiJiraRequestGuideContent.model_fields)),
    )
    with open_session(get_ai_postgres_engine()) as session:
        session.merge(ai_request_guide)
        for field_name, description in state.input_request_fields.model_dump().items():
            field = AiInputRequestFields(
                **index,
                field_id=fields.inverse[field_name],
                description=description,
            )
            session.merge(field)

        session.commit()

    update_ai_jira_request_guide_embedding(ai_request_guide)
    return ai_request_guide


def build_summarization_context_prompt(
    service_desk_id: int, request_type_id: int, input_attributes: list[str]
) -> str:
    service_desk_name = get_service_desks(get_service_desk())[service_desk_id]["projectName"]
    request_type_name = get_service_desks_request_types(get_service_desk(), service_desk_id)[
        request_type_id
    ]["name"]
    context = {
        "Organization": (
            "Kapital Bank, a bank in Azerbaijan servicing both retail and corporate customers, "
            "providing services both at physical branches and digitally online."
        ),
        "Jira Service Desk name": service_desk_name,
        "Request type": request_type_name,
        "Request attributes that were provided by users": input_attributes,
    }
    prompt = textwrap.dedent("""
        I work at an organization where we use a Jira service desk to manage various requests. 
        Help me analyze resolved requests of one specific type from this service desk to create a guide for new users. 
        This guide should explain when to use this request type (the situations or problems it solves) 
        and how to correctly fill in information for this request type.

        You will be provided with a **JSON** list containing multiple **resolved requests** of the same request type.
        Each item in the list corresponds to an individual request with same attributes for all requests.
        All attribute values were **provided by users** when creating the requests, except for the following attributes:
        - **Resolution**: Indicates the outcome of the request, where "Done" and "Resolved" signifies a successful resolution, 
            while "Rejected", "Won't Do", "Declined" and "Unresolved" usually means the request was incorrectly formulated or not approved.
        - **Comments**: Contains the **discussion history** and any communication between the request participants during the resolution process. 
            The format is a list of strings with comment author in the prefix, for example:
            ```json
            [
                "Reporter: Please help me change the password",
                "Assignee: Hello, the link was sent in email",
                "Reporter: Thank you",
            ]
            ```

        ### Your Task
        1. **Analyze the provided requests** to determine:
        - **What kind of problems this request type solves** (e.g., customer issues, operational tasks, or system-related inquiries).
        - **What information a user needs to provide** to ensure the request is accurately processed and resolved. 
            For each attribute field that the user needs to provide, generate a short description and explain why this attribute is needed 
            and how it affects request resolution.
        - **How the request is resolved** - the steps typically taken by both the **reporter** and **assignee** to resolve these requests.

        2. **Language consideration**:
        - Be aware that most of the data will be in **Azerbaijani**. Interpret the content accordingly while identifying patterns and extracting 
            relevant information. But note that the output should be in **English**.

        ### Additional Context
        Additional information regarding the environment in which the requests were created, meant to clarify references used in requests 
        and aid in creating an informed user guide:
    """) + to_json_llm_input(context)  # noqa: E501
    return prompt


def update_ai_jira_request_guide_embedding(guide: AiJiraRequestGuide):
    metadata = guide.model_dump(include=get_index_column_names())
    document = langchain_core.documents.Document(
        page_content=guide.use_case,
        metadata=metadata,
    )
    store = get_postgres_vector_store()
    existing_documents_with_same_metadata = store.similarity_search_by_vector(
        [0] * store.embeddings.dimensions, k=1000, filter=metadata
    )
    store.delete([d.id for d in existing_documents_with_same_metadata])
    store.add_documents([document])


def update_all_ai_jira_request_guide_embeddings():
    with open_session(get_ai_postgres_engine()) as session:
        statement = sqlmodel.select(AiJiraRequestGuide)
        for guide in tqdm.auto.tqdm(session.exec(statement), desc="Updating embeddings"):
            update_ai_jira_request_guide_embedding(guide)



import dataclasses
import enum
from typing import Iterable

import numpy as np

from ai_app.utils import int_log_2


class ModelProvider(enum.StrEnum):
    openai = enum.auto()
    google_genai = enum.auto()
    azure_openai = enum.auto()


@dataclasses.dataclass
class ModelMetaAttributes:
    name: str
    provider: ModelProvider
    million_input_tokens_cost: float
    million_output_tokens_cost: float
    context_window: int
    supports_system_messages: bool

    @property
    def python_name(self):
        name = self.name.replace("-", "_").replace(".", "_")
        return name

    def get_weighted_cost(self, input_token_weight: float = 10.0):
        cost = np.average(
            [self.million_input_tokens_cost, self.million_output_tokens_cost],
            weights=[input_token_weight, 1],
        )
        return cost

    def get_relative_cost(self, input_token_weight: float = 10.0, lowest_cost=0.25) -> int:
        """
        Returns a relative integer cost >= 1, with each subsequent int signifying a double increase
        in cost.
        """
        cost = self.get_weighted_cost(input_token_weight)
        cost = int_log_2(cost)
        lowest_cost = int_log_2(lowest_cost)
        cost = 1 + max(0, cost - lowest_cost)
        return cost

    def get_relative_cost_representation(self, **kwargs) -> str:
        cost = "💲" * self.get_relative_cost(**kwargs)
        return cost


def get_common_models_meta_attributes() -> dict[str, ModelMetaAttributes]:
    """
    Note that cost depends on token type (input, output, reasoning)
    and on input/output type (text, audio, image, video).
    Pricing is relevant as of 2025.04.23.
    """
    models = [
        ("gemini-2.5-flash", "google_genai", 0.30, 2.50, 1_000_000, True),
        ("gemini-2.5-pro", "google_genai", 1.25, 10.00, 2_000_000, True),
        ("gpt-5", "openai", 1.25, 10.00, 400_000, True),
        ("gpt-5-mini", "openai", 0.25, 2.00, 400_000, True),
        ("gpt-5-nano", "openai", 0.05, 0.40, 400_000, True),
        ("gpt-4.1", "openai", 2.00, 8.00, 1_000_000, True),
        ("gpt-4.1-mini", "openai", 0.40, 1.60, 1_000_000, True),
        ("gpt-4.1-nano", "openai", 0.10, 0.40, 1_000_000, True),
        ("o4-mini", "openai", 1.10, 4.40, 200_000, True),
        ("o3", "openai", 2, 8, 200_000, False),
    ]
    models = [ModelMetaAttributes(*a) for a in models]
    models = sorted(models, key=lambda model: model.get_weighted_cost())
    models = {m.name: m for m in models}
    return models


# ModelName = enum.StrEnum(
#     "ModelName",
#     names=[m.python_name for m in get_common_models_meta_attributes().values()],
# )


def filter_common_model_names(
    max_cost: float = float("inf"),
    supports_system_messages: bool = False,
    providers: Iterable[ModelProvider] | None = None,
) -> list[str]:
    model_names = []
    for name, model in get_common_models_meta_attributes().items():
        if model.get_weighted_cost() > max_cost:
            continue
        if providers and model.provider not in providers:
            continue
        if supports_system_messages and not model.supports_system_messages:
            continue

        model_names.append(name)

    return model_names



import enum
import json
import logging
import re
import urllib
import warnings

import atlassian
import bidict
import polars as pl
import tqdm.auto

from ai_app.utils import filter_dict_by_keys, get_number_of_batches


def get_jira_base_url() -> str:
    return "http://jira-support.kapitalbank.az"


def fetch_all_jira_project_keys(jira: atlassian.Jira) -> list[str]:
    projects = jira.get_all_projects()
    projects = pl.DataFrame(projects, infer_schema_length=None)
    project_keys = sorted(set(projects["key"]))
    return project_keys


def fetch_all_jira_usernames(jira: atlassian.Jira) -> list[str]:
    start = 0
    users = []
    while True:
        batch_users = jira.user_find_by_user_string(username=".", start=start, limit=500)
        if not batch_users:
            break

        start += len(batch_users)
        users.extend(batch_users)

    users = pl.DataFrame(users, infer_schema_length=None)
    usernames = sorted(set(users["name"]))
    return usernames


def build_issue_url(issue_key: str, base_url: str | None = None) -> str:
    base_url = base_url or get_jira_base_url()
    issue_url = f"{base_url}/browse/{issue_key}"
    return issue_url


class IssueLink(enum.StrEnum):
    tests = "Tests"
    has_to_be_done_before = "Has to be done before"

    def get_link_type(self):
        link_types = {
            IssueLink.has_to_be_done_before: "Gantt End to Start",
            IssueLink.tests: "Tests",
        }
        link_type = link_types[self]
        return link_type

    def link_issues(self, jira: atlassian.Jira, inward_issue: str, outward_issue: str) -> None:
        jira.create_issue_link(
            {
                "type": {"name": self.get_link_type()},
                "inwardIssue": {"key": inward_issue},
                "outwardIssue": {"key": outward_issue},
            }
        )


def extract_jira_issue_fields(issue: dict, max_length: int | None = None) -> dict[str, str]:
    extracted_fields = {}
    for key, value in issue["fields"].items():
        match key:
            case "issuetype":
                value = value["name"]

        value = value[:max_length] if max_length and value else value
        extracted_fields[key] = value

    return extracted_fields


def get_service_desks(service_desk: atlassian.ServiceDesk) -> dict:
    service_desks = service_desk.get_service_desks()
    service_desks = {int(i["id"]): i for i in service_desks}
    return service_desks


def generate_from_pages(
    service_desk: atlassian.ServiceDesk,
    path: str,
    limit: int | None = None,
    verbose: bool = False,
    tqdm_description="Fetching pages",
    **kwargs,
):
    progress_bar = tqdm.auto.tqdm(desc=tqdm_description, disable=not verbose)
    start = 0
    while True:
        response = service_desk.get(path, params=dict(start=start), **kwargs)
        start += response["size"]
        yield from response["values"]
        progress_bar.update()
        if response["isLastPage"] or limit and start > limit:
            progress_bar.close()
            return


def get_service_desks_request_types(
    service_desk: atlassian.ServiceDesk,
    service_desk_id: int,
    tqdm_description="Fetching request types",
    verbose: bool = False,
) -> dict:
    """Collect all request types for given service desk."""
    request_types = generate_from_pages(
        service_desk,
        f"/rest/servicedeskapi/servicedesk/{service_desk_id}/requesttype",
        tqdm_description=tqdm_description,
        verbose=verbose,
    )
    request_types = {int(i["id"]): i for i in request_types}
    return request_types


def build_request_url(
    service_desk_id: int, request_type_id: int, base_url: str | None = None, **parameters
) -> str:
    base_url = base_url or get_jira_base_url()
    url = f"{base_url}/servicedesk/customer/portal/{service_desk_id}/create/{request_type_id}"
    if parameters:
        query = urllib.parse.urlencode(parameters)
        url += f"?{query}"

    return url


def generate_jql_issues(
    jira: atlassian.Jira,
    jql: str,
    verbose: bool = False,
    fields: list[str] | str | None = None,
    expand: list[str] | str | None = None,
    soft_limit: int | None = None,
):
    def to_expected_jira_format(value: list[str] | str | None = None):
        value = value if value is None or isinstance(value, str) else ",".join(value)
        return value

    progress_bar = tqdm.auto.tqdm(desc="Fetching issue pages", disable=not verbose)
    fetched_issues_count = 0
    soft_limit = soft_limit or float("inf")
    while fetched_issues_count < soft_limit:
        try:
            response = jira.jql(
                jql,
                start=fetched_issues_count,
                fields=to_expected_jira_format(fields),
                expand=to_expected_jira_format(expand),
                limit=min(soft_limit, 50),
            )
        except Exception:
            logging.error(f"Failed to fetch issues from JQL: {jql}")
            raise

        fetched_issues_count += len(response["issues"])
        soft_limit = int(min(soft_limit, response["total"]))
        if not progress_bar.total:
            progress_bar.total = get_number_of_batches(soft_limit, response["maxResults"])

        progress_bar.update()
        yield from response["issues"]

    progress_bar.close()


def normalize_line_breaks(string: str) -> str:
    string = string.replace("\r\n", "\n")
    string = re.sub(r"\n+", r"\n", string)
    return string


def try_inferring_jira_field_representation(
    value, dict_key_preference=("displayName", "name", "value", "key")
):
    def recursive_infer(value):
        if isinstance(value, list):
            value = [recursive_infer(i) for i in value]
            if all(isinstance(i, str) for i in value):
                value = ", ".join(value)

        if isinstance(value, dict):
            for key in dict_key_preference:
                if key in value:
                    value = value[key]
                    value = recursive_infer(value)
                    return value

        return value

    value = recursive_infer(value)
    return value


def extract_issue_field_values(issue: dict, include_issue_key: bool = True) -> dict:
    values = {}
    if include_issue_key:
        values["Issue key"] = issue["key"]

    for field, value in issue["fields"].items():
        if not value:
            values[field] = None
            continue

        match field:
            case "issuetype" | "priority" | "status" | "resolution":
                value = value["name"]
            case (
                "customfield_20163"
                | "customfield_20901"
                | "customfield_20902"
                | "customfield_20906"
                | "customfield_23001"
                | "customfield_23101"
                | "customfield_23608"
                | "customfield_23609"
                | "customfield_23710"
                | "customfield_23712"
                | "customfield_15542"
                | "customfield_24810"
                | "customfield_23109"
            ):
                value = value["value"]
            case "project":
                value = value["key"]
            case (
                "labels"
                | "customfield_20160"
                | "customfield_20800"
                | "customfield_20164"
                | "customfield_20118"
                | "customfield_11626"
                | "customfield_11300"
            ):
                value = ", ".join(value)
            case "components" | "customfield_20145" | "customfield_10300":
                value = ", ".join(i["name"] for i in value)
            case "assignee" | "reporter":
                value = value["displayName"]
            case "comment":
                value = sorted(value["comments"], key=lambda c: c["created"])
                value = [c["body"] for c in value]
                value = json.dumps(value)

        value = try_inferring_jira_field_representation(value)

        valid_types = [type(None), int, float, str]
        if not any(isinstance(value, t) for t in valid_types):
            raise ValueError(
                f"Processed field '{field}' has unexpected type {type(value)}:\n{value}"
            )

        if isinstance(value, str):
            value = normalize_line_breaks(value)

        values[field] = value

    return values


def get_request_fields(
    service_desk: atlassian.ServiceDesk, service_desk_id: int, request_type_id: int
) -> pl.DataFrame | None:
    response = service_desk.get(
        path=f"/rest/servicedeskapi/servicedesk/{service_desk_id}/requesttype/{request_type_id}/field",
    )
    df = pl.DataFrame(response["requestTypeFields"])
    if df.is_empty():
        return

    can_be_passed_as_url_parameter = {
        "any": False,
        "array": False,
        "date": True,  # in format 1/Apr/25, https://jira-support.kapitalbank.az/servicedesk/customer/portal/21/create/2495
        "datetime": True,  # in format 1/Apr/25 09:56 AM, https://jira-support.kapitalbank.az/servicedesk/customer/portal/21/create/2495
        "option": True,
        "priority": True,  # similar to option, https://jira-support.kapitalbank.az/servicedesk/customer/portal/21/create/5155
        "string": True,
        "user": False,  # similar to any, https://jira-support.kapitalbank.az/servicedesk/customer/portal/21/create/2018
        "number": True,  # just string with frontend/backend validation https://jira-support.kapitalbank.az/servicedesk/customer/portal/1/create/2287
        "BundledFields": False,  # Kampaniyanın keçirilmə aralığı in https://jira-support.kapitalbank.az/servicedesk/customer/portal/1/create/1838
    }
    field_type_expr = pl.col("jiraSchema").struct["type"]
    if field_types := (set(df.select(field_type_expr.unique()).to_series())) - set(
        can_be_passed_as_url_parameter
    ):
        warnings.warn(
            f"Unexpected field types: {field_types} for service desk {service_desk_id} "
            f"and request type {request_type_id}\n"
            f"Request URL: {build_request_url(service_desk_id, request_type_id)}"
        )

    df = df.with_columns(
        field_type_expr.replace_strict(can_be_passed_as_url_parameter, default=False).alias(
            "can_be_passed_as_url_parameter"
        ),
    )
    try:
        df = df.sort("fieldId").with_columns(
            pl.col("validValues")
            .explode()
            .struct["label"]
            .drop_nulls()
            .implode()
            .over("fieldId", order_by="fieldId")
            .alias("valid_value_labels")
        )
    except pl.exceptions.StructFieldNotFoundError:
        df = df.with_columns(pl.lit(None, dtype=list[str]).alias("valid_value_labels"))

    df = df.with_columns(
        pl.when(pl.col("valid_value_labels").list.len() > 0).then("valid_value_labels")
    )
    return df


def get_request_field_names(
    service_desk: atlassian.ServiceDesk, service_desk_id: int, request_type_id: int
) -> bidict.bidict:
    """Returns field id to name bidict, keeps the largest field id if names are duplicated."""
    request_field_names = bidict.bidict()
    request_fields = get_request_fields(service_desk, service_desk_id, request_type_id)
    if request_fields is None:
        return request_field_names

    request_fields = request_fields.sort("fieldId").unique("name", keep="last")
    request_field_names.putall(
        (row["fieldId"], row["name"]) for row in request_fields.iter_rows(named=True)
    )
    return request_field_names


def build_new_request_url_similar_to_given(
    jira: atlassian.Jira, service_desk: atlassian.ServiceDesk, issue_key: str
) -> str:
    """Note that partial url parameters don't work for fields which are linked to Jira assets."""
    issue = jira.issue(issue_key)
    request_type = issue["fields"]["customfield_10202"]["requestType"]
    service_desk_id = int(request_type["serviceDeskId"])
    request_type_id = int(request_type["id"])
    fields = get_request_fields(service_desk, service_desk_id, request_type_id)
    fields = list(fields["fieldId"])
    fields = {f: issue["fields"][f] for f in fields}
    fields = {
        k: try_inferring_jira_field_representation(
            v, dict_key_preference=["displayName", "id", "name", "value", "key"]
        )
        for k, v in fields.items()
    }
    url = build_request_url(service_desk_id, request_type_id, **fields)
    return url


def build_jql_last_resolved(project_key: str) -> str:
    """Latest resolved issues, ordered by resolution time descending."""
    return f"project = {project_key} AND resolutiondate IS NOT EMPTY ORDER BY resolutiondate DESC"


def build_jql_current_open(project_key: str) -> str:
    """Currently open issues. Using resolution = Unresolved is robust across workflows."""
    return f"project = {project_key} AND resolution = Unresolved ORDER BY created DESC"


def get_project_overview(
    jira: atlassian.Jira,
    project_key: str,
    resolved_issues_count: int = 10,
    unresolved_issues_count: int = 10,
    issue_fields: list[str] = (
        "summary",
        "description",
        "issuetype",
        "priority",
        "comment",
    ),
) -> dict:
    # TODO: Add project boards, add fields:
    # "issuelinks",
    # "parent",
    # "subtasks",
    #  Epic links
    project = jira.project(project_key, expand="description")
    project = filter_dict_by_keys(project, ["key", "name", "description"])
    jqls = {
        "resolved": (
            f"""
                project = {project_key} AND resolutiondate IS NOT EMPTY
                ORDER BY resolutiondate DESC
            """,
            resolved_issues_count,
        ),
        "unresolved": (
            f"""
                project = {project_key} AND resolution = Unresolved
                ORDER BY created DESC
            """,
            unresolved_issues_count,
        ),
    }
    for kind, (jql, limit) in jqls.items():
        issues = []
        for issue in generate_jql_issues(jira, jql, fields=issue_fields, soft_limit=limit):
            if comments := issue["fields"].pop("comment", None):
                comments = sorted(comments["comments"], key=lambda c: c["created"])
                comments = [c["body"] for c in comments]

            issue = extract_issue_field_values(issue)
            issue["comments"] = comments
            issues.append(issue)

        project[f"{kind}_issues"] = issues

    return project



from typing import override

import gradio as gr
import pandas as pd

from ai_app.core import BaseApp
from ai_app.frontend import build_model_choice_dropdown

from .backend import SpecificationEvaluator


def prettify(col: str) -> str:
    col = col.replace("_", " ")
    col = col.capitalize()
    return col


class App(BaseApp):
    name = "Evaluate software requirements specification"
    requires_auth = False

    def __init__(self):
        self.specification_evaluator = SpecificationEvaluator()

    def evaluate_software_requirements_specification(self, model_name: str, specification: str):
        evaluation, run_id = self.specification_evaluator.evaluate(
            model_name=model_name, specification=specification
        )
        characteristics = evaluation.characteristics.model_dump()
        characteristics = [
            {
                "Characteristic": prettify(k),
                "Grade": v["grade"],
                "Feedback": v["feedback"],
            }
            for k, v in characteristics.items()
        ]
        characteristics = pd.DataFrame(characteristics)
        characteristics["Grade"] = characteristics["Grade"].clip(1, 5)
        outputs = [
            characteristics,
            evaluation.specification_improvement_recommendation,
            evaluation.specification_improvement_example,
        ]
        return outputs

    @override
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks() as app:
            model_choice = model_choice or build_model_choice_dropdown()

            gr.Markdown(
                "This app analyzes software requirement specifications based on the IEEE standard and provides recommendations for improvement"
            )
            specification = gr.TextArea(
                label="Software requirements specification",
                # Remove the value in production.
                value="The App class should implement a range of methods that the frontend is going to call, such as GetCurrentBalance, which takes a timestamp and returns the current balance on all the user accounts.",
            )
            evaluate = gr.Button("Evaluate", variant="primary")
            characteristics = gr.BarPlot(
                pd.DataFrame(),
                label="Desirable specification characteristics",
                x="Characteristic",
                y="Grade",
                y_lim=(0, 5),
                tooltip=["Feedback"],
                title="Hover on a characteristic bar to see detailed feedback",
            )
            recommendation = gr.TextArea(label="Improvement recommendation")
            example = gr.TextArea(label="Improvement example")

            evaluate.click(
                self.evaluate_software_requirements_specification,
                inputs=[model_choice, specification],
                outputs=[characteristics, recommendation, example],
            )

        return app



"""
Quickstart: https://developers.google.com/sheets/api/quickstart/python
Python client: https://github.com/googleapis/google-api-python-client
Using service account: https://github.com/googleapis/google-api-python-client/blob/main/docs/oauth-server.md

1. Create Google cloud project
2. Enable necessary API, IAM

"""

import json
import mimetypes
import pathlib

import google.cloud.iam_admin_v1
import google.oauth2.credentials
import google.oauth2.service_account
import google_auth_oauthlib
import googleapiclient.discovery
import googleapiclient.http
import polars as pl


def create_service_account_info(project: str, service_account: str) -> dict:
    """A service account can be built in the Google cloud console."""
    iam_admin_client = google.cloud.iam_admin_v1.IAMClient()
    request = google.cloud.iam_admin_v1.types.CreateServiceAccountKeyRequest()
    request.name = f"projects/{project}/serviceAccounts/{service_account}"
    key = iam_admin_client.create_service_account_key(request=request)
    info = json.loads(key.private_key_data)
    return info


def create_authorized_user_info(user_credentials: dict, scopes: list[str]) -> dict:
    """Authorized user credentials can be downloaded from the Google cloud console."""
    flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_config(user_credentials, scopes)
    credentials = flow.run_local_server(port=0)
    info = json.loads(credentials.to_json())
    return info


def get_credentials(
    *,
    service_account_info: dict | None = None,
    authorized_user_info: dict | None = None,
    scopes: list[str],
) -> google.auth.credentials.Credentials:
    if bool(service_account_info) == bool(authorized_user_info):
        raise ValueError(
            "Exactly one of service_account_info and authorized_user_info must be provided."
        )

    if service_account_info:
        credentials = google.oauth2.service_account.Credentials.from_service_account_info(
            service_account_info
        )
        credentials = credentials.with_scopes(scopes)
    else:
        credentials = google.oauth2.credentials.Credentials.from_authorized_user_info(
            authorized_user_info, scopes
        )

    if credentials.expired:
        credentials.refresh(google.auth.transport.requests.Request())

    return credentials


def build_spreadsheets_api(credentials: google.auth.credentials.Credentials, version: str = "v4"):
    resource = googleapiclient.discovery.build("sheets", version, credentials=credentials)
    spreadsheets_api = resource.spreadsheets()
    return spreadsheets_api


class Drive:
    def __init__(self, credentials: google.auth.credentials.Credentials, version: str = "v3"):
        self.api = googleapiclient.discovery.build("drive", version, credentials=credentials)

    def upload(
        self,
        filepath: str,
        name: str | None = None,
        mime_type: str | None = None,
        make_public: bool = False,
        parents: list[str] | None = None,
    ) -> str:
        """Upload a file to Drive and return a shareable URL.

        - name: optional override for the file name shown in Drive.
        - mime_type: if not provided, it will be guessed from the file extension.
        - make_public: if True, sets the file permission to anyone-with-link reader.
        - parents: list of folder IDs to upload into.
        """

        path = pathlib.Path(filepath)
        if not path.exists():
            raise FileNotFoundError(f"File not found: {filepath}")

        guessed_mime = mime_type or (
            mimetypes.guess_type(str(path))[0] or "application/octet-stream"
        )
        media = googleapiclient.http.MediaFileUpload(
            str(path), mimetype=guessed_mime, resumable=False
        )

        body: dict = {"name": name or path.name}
        if parents:
            body["parents"] = parents

        created = (
            self.api.files()
            .create(body=body, media_body=media, fields="id,webViewLink,webContentLink")
            .execute()
        )

        file_id = created.get("id")
        if not file_id:
            raise RuntimeError("Drive API did not return a file id")

        if make_public:
            # Best-effort: allow anyone with the link to view
            self.api.permissions().create(
                fileId=file_id,
                body={"type": "anyone", "role": "reader"},
            ).execute()

        # Prefer webViewLink; fallback to deterministic viewer URL
        url = created.get("webViewLink") or f"https://drive.google.com/file/d/{file_id}/view"
        return url


class Spreadsheets:
    @classmethod
    def create(cls, name: str, credentials: google.auth.credentials.Credentials):
        api = build_spreadsheets_api(credentials)
        spreadsheet = api.create(body={"properties": {"title": name}}).execute()
        spreadsheet = cls(spreadsheet["spreadsheetId"], credentials)
        return spreadsheet

    def __init__(self, spreadsheet_id: str, credentials: google.auth.credentials.Credentials):
        self.id = spreadsheet_id
        self.api = build_spreadsheets_api(credentials)

    def build_range(self, sheet_name: str, start: str = "A1", end: str = "ZZZ") -> str:
        return f"'{sheet_name}'!{start}:{end}"

    def read_sheet(self, sheet_name: str, with_header: bool = True) -> pl.DataFrame | None:
        sheet = (
            self.api.values()
            .get(spreadsheetId=self.id, range=self.build_range(sheet_name))
            .execute()
        )
        if "values" not in sheet:
            return

        n_columns = max(len(row) for row in sheet["values"])
        if with_header:
            columns, *data = sheet["values"]
        else:
            columns = None
            data = sheet["values"]

        data = [(row + [None] * n_columns)[:n_columns] for row in data]
        df = pl.DataFrame(data, columns, orient="row", infer_schema_length=None)
        return df

    def exists(self, sheet_name: str) -> bool:
        try:
            self.read_sheet(sheet_name)
            return True
        except googleapiclient.errors.HttpError:
            return False

    def clear_sheet(self, sheet_name: str):
        self.api.values().clear(
            spreadsheetId=self.id,
            range=self.build_range(sheet_name),
        ).execute()

    def write_sheet(self, sheet_name: str, df: pl.DataFrame):
        if self.exists(sheet_name):
            self.clear_sheet(sheet_name)
        else:
            self.create_sheet(sheet_name)

        df = df.fill_nan("#N/A")
        values = [df.columns] + list(df.iter_rows())
        self.api.values().update(
            spreadsheetId=self.id,
            range=self.build_range(sheet_name),
            valueInputOption="USER_ENTERED",
            body={"values": values},
        ).execute()

    def describe(self) -> pl.DataFrame:
        response = self.api.get(spreadsheetId=self.id).execute()
        df = pl.DataFrame(response["sheets"], strict=False)
        df = df.unnest("properties").unnest("gridProperties")
        return df

    @property
    def title_to_id(self) -> dict:
        return dict(self.describe()[["title", "sheetId"]].iter_rows())

    def to_sheet_id(self, sheet: int | str) -> int:
        return sheet if isinstance(sheet, int) else self.title_to_id[sheet]

    def _batch_update(self, *requests: dict) -> dict:
        response = self.api.batchUpdate(
            spreadsheetId=self.id, body={"requests": requests}
        ).execute()
        return response

    def create_sheet(self, name: str) -> str:
        request = {"addSheet": {"properties": {"title": name}}}
        sheet = self._batch_update(request)
        sheet_id = sheet["replies"][0]["addSheet"]["properties"]["sheetId"]
        return sheet_id

    def delete_sheet(self, sheet: int | str):
        request = {"deleteSheet": {"sheetId": self.to_sheet_id(sheet)}}
        self._batch_update(request)

    def copy_sheet(self, sheet: int | str) -> int:
        sheet = (
            self.api.sheets()
            .copyTo(
                spreadsheetId=self.id,
                sheetId=self.to_sheet_id(sheet),
                body=dict(destinationSpreadsheetId=self.id),
            )
            .execute()
        )
        sheet_id = sheet["sheetId"]
        return sheet_id

    def rename_sheet(self, sheet: int | str, name: str):
        request = {
            "updateSheetProperties": {
                "properties": {
                    "sheetId": self.to_sheet_id(sheet),
                    "title": name,
                },
                "fields": "title",
            }
        }
        self._batch_update(request)

    def copy_and_write(
        self, copy_from: str, copy_to: str, df: pl.DataFrame, overwrite: bool = False
    ):
        sheet_id = self.copy_sheet(copy_from)
        if overwrite and copy_to in self.describe()["title"]:
            self.delete_sheet(copy_to)

        self.rename_sheet(sheet_id, copy_to)
        self.write_sheet(copy_to, df)


def get_google_sheets_credentials_scopes():
    return ["https://www.googleapis.com/auth/spreadsheets"]


def get_google_drive_credentials_scopes():
    return ["https://www.googleapis.com/auth/drive.file"]


def build_spreadsheets(spreadsheet_id: str, service_account_info: str) -> Spreadsheets:
    service_account_info = json.loads(service_account_info)
    credentials = get_credentials(
        service_account_info=service_account_info,
        scopes=get_google_sheets_credentials_scopes(),
    )
    spreadsheets = Spreadsheets(spreadsheet_id=spreadsheet_id, credentials=credentials)
    return spreadsheets


def build_drive(authorized_user_info: str) -> Drive:
    authorized_user_info = json.loads(authorized_user_info)
    credentials = google.oauth2.credentials.Credentials.from_authorized_user_info(
        authorized_user_info, scopes=get_google_drive_credentials_scopes()
    )
    if credentials.expired:
        if credentials.refresh_token:
            credentials.refresh(google.auth.transport.requests.Request())
        else:
            raise ValueError("Credentials are expired and have no refresh token.")

    drive = Drive(credentials=credentials)
    return drive


def run_interactive_user_credentials_flow(client_secrets_path: str, scopes: list[str]):
    """
    The client_secrets can be downloaded from the Google cloud console.
    The resulting credentials can be saved as authorized_user_info via creds.to_json().
    """
    flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_secrets_file(
        client_secrets_path, scopes
    )
    credentials = flow.run_local_server(port=0)
    return credentials



import inspect
import uuid
from abc import ABC, abstractmethod
from typing import AsyncGenerator, Self

import gradio as gr
import langchain_core.language_models
import langchain_core.messages
import langgraph.errors
import langsmith
import pydantic

from ai_app.config import get_anon_codec, get_chat_model, get_config
from ai_app.frontend import build_model_choice_dropdown
from ai_app.pii import EntityTypeCodec
from ai_app.utils import PydanticForbidExtra


class State(PydanticForbidExtra, arbitrary_types_allowed=True):
    thread_id: uuid.UUID = pydantic.Field(default_factory=uuid.uuid4)
    name: str | None = None
    email: str | None = None
    messages: list[dict] = []
    codec: EntityTypeCodec = pydantic.Field(
        default_factory=lambda: get_config().build_entity_codec()
    )

    def copy_for_new_conversation(self) -> Self:
        state = State(**self.model_dump(include=["name", "email"]))
        return state

    @property
    def langsmith_extra(self) -> dict:
        langsmith_extra = {"metadata": {"thread_id": self.thread_id}}
        return langsmith_extra

    @classmethod
    def from_request(cls, request: gr.Request) -> Self:
        session = request.request.session
        if user_info := session.get("user_info"):
            state = cls(
                name=user_info.get("name"),
                email=user_info.get("email"),
            )
        else:
            state = cls()

        return state


class Response(pydantic.BaseModel):
    content: str
    additional_outputs: list = []
    messages: list = []
    content_for_langsmith: str | None = pydantic.Field(
        default=None,
        description=inspect.cleandoc("""
            The message to show as agent response in LangSmith UI, 
            which can't render HTML, for example.
        """),
    )


class BaseApp(ABC):
    name: str
    requires_auth: bool

    @abstractmethod
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks: ...

    @classmethod
    def get_api_prefix(cls) -> str:
        api_prefix = cls.name.replace(" ", "_").lower()
        return api_prefix

    @classmethod
    def get_chat_api_name(cls, absolute: bool = False) -> str:
        namespace = "/" if absolute else ""
        api_name = f"{namespace}{cls.get_api_prefix()}_chat"
        return api_name

    async def _respond(
        self,
        message: str,
        conversation: list,
        model_name: str,
        state: State,
        *args,
        **kwargs,
    ):
        # If user cleared the chat or switched to new one in the same interface, need to reset the
        # messages in state also. This would not be necessary if the frontend messages were the
        # single source of truth.
        if not conversation:
            state = state.copy_for_new_conversation()

        model = get_chat_model(model_name)
        message = get_anon_codec().anonymize(message, codec=state.codec)
        message = langchain_core.messages.HumanMessage(message)
        frontend_messages = langchain_core.messages.convert_to_messages(conversation)
        trace = langsmith.trace(
            name=self.name,
            inputs={"message": message},
            metadata={
                "thread_id": state.thread_id,
                "user_name": state.name,
                "user_email": state.email,
            },
        )
        async with trace as run:
            try:
                async for response in self.astream(
                    model, message, frontend_messages, state, *args, **kwargs
                ):
                    state.messages = response.messages
                    yield [response.content] + response.additional_outputs + [state]
            except langgraph.errors.GraphRecursionError:
                raise gr.Error("AI agent exceeded maximum recursion depth.")

            run.add_outputs({"response": response.content_for_langsmith or response.content})

    async def astream(
        self,
        model: langchain_core.language_models.BaseChatModel,
        message: langchain_core.messages.BaseMessage,
        frontend_messages: list[langchain_core.messages.BaseMessage],
        state: State,
        *args,
        **kwargs,
    ) -> AsyncGenerator[Response, None]:
        """Need to be implemented only for apps with chat interface."""
        yield await self.respond(model, message, frontend_messages, state, *args, **kwargs)

    async def respond(
        self,
        model: langchain_core.language_models.BaseChatModel,
        message: langchain_core.messages.BaseMessage,
        frontend_messages: list[langchain_core.messages.BaseMessage],
        state: State,
        *args,
        **kwargs,
    ) -> Response:
        """Need to be implemented only for apps with chat interface."""
        raise NotImplementedError

    def build_gradio_chat_interface(
        self,
        model_choice: gr.Dropdown | None = None,
        additional_inputs: list | None = None,
        additional_outputs: list | None = None,
        **kwargs,
    ) -> gr.ChatInterface:
        model_choice = model_choice or build_model_choice_dropdown()
        self.state = gr.State(State())
        chat_interface = gr.ChatInterface(
            fn=self._respond,
            type="messages",
            additional_inputs=[model_choice, self.state] + (additional_inputs or []),
            additional_outputs=(additional_outputs or []) + [self.state],
            api_name=self.get_chat_api_name(),
            **get_config().common_chat_interface_parameters,
            **kwargs,
        )
        chat_interface.load(State.from_request, outputs=self.state)
        return chat_interface



import langchain.chat_models
import langchain_core
import pydantic

from ai_app.ai_utils import StructuredOutputMethod
from ai_app.external.providers import ModelProvider
from ai_app.utils import PydanticForbidExtra


class ChatModelParameters(PydanticForbidExtra):
    model: str
    model_provider: ModelProvider | None = None
    api_key: pydantic.SecretStr | None = None
    preferred_structured_output_method: StructuredOutputMethod | None = None
    max_possible_completion_tokens: int | None = None

    @classmethod
    def from_model_name(cls, model: str):
        structured_output_method = None
        max_possible_completion_tokens = None
        provider = None
        if model.startswith("gemini"):
            provider = ModelProvider.google_genai
            if model == "gemini-1.5-flash":
                max_possible_completion_tokens = 4096

        elif model.startswith("gpt") or model.startswith("o"):
            provider = ModelProvider.openai
            if model.startswith("gpt-3"):
                structured_output_method = StructuredOutputMethod.function_calling
            else:
                structured_output_method = StructuredOutputMethod.json_schema

            if model == "gpt-4o-mini":
                max_possible_completion_tokens = 16_384

        parameters = cls(
            model_provider=provider,
            model=model,
            preferred_structured_output_method=structured_output_method,
            max_possible_completion_tokens=max_possible_completion_tokens,
        )
        return parameters

    def build_model(
        self, schema=None, max_tokens: int | None = None, **kwargs
    ) -> langchain_core.language_models.BaseChatModel:
        max_tokens = max_tokens or self.max_possible_completion_tokens
        if self.max_possible_completion_tokens:
            max_tokens = min(max_tokens, self.max_possible_completion_tokens)

        model = langchain.chat_models.init_chat_model(
            model=self.model,
            model_provider=self.model_provider,
            api_key=self.api_key,
            max_tokens=max_tokens,
            **kwargs,
        )
        if schema:
            if self.preferred_structured_output_method:
                model = model.with_structured_output(
                    schema, method=self.preferred_structured_output_method
                )
            else:
                model = model.with_structured_output(schema)

        return model



from typing import override

import bir_mcp
import langchain.tools

from ai_app.tools.base import BaseToolkit


class SonarQubeToolkit(BaseToolkit):
    def __init__(
        self,
        url: str,
        token: str,
        gitlab_url: str,
        timezone: str = "UTC",
        max_tool_output_length: int | None = None,
    ):
        super().__init__()
        self.mcp = bir_mcp.git_lab.SonarQube(
            url=url, gitlab_url=gitlab_url, token=token, timezone=timezone
        )
        self.max_tool_output_length = max_tool_output_length

    @override
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        tools = [
            self.mcp.get_project_quality_gates_status,
            self.mcp.get_main_project_metrics,
        ]
        tools = [
            bir_mcp.utils.to_langchain_tool(
                tool, prefix=self.mcp.get_tag(), max_output_length=self.max_tool_output_length
            )
            for tool in tools
        ]
        return tools



import inspect

import langchain_core.messages
import langgraph.graph
import langgraph.prebuilt
import pydantic

from ai_app.apps.jira_service_desk.tools import (
    JiraServiceDeskToolkit,
    fetch_detailed_jira_request_type_guide,
)
from ai_app.tools import build_structured_response_tool, think


async def ainvoke_model_with_tools(
    model,
    tools,
    messages: list[langchain_core.messages.BaseMessage],
    system_prompt: str | None = None,
    tool_choice: str = "any",
) -> list[langchain_core.messages.AIMessage | langchain_core.messages.ToolMessage]:
    model = model.bind_tools(tools, tool_choice=tool_choice)
    if system_prompt:
        messages = [langchain_core.messages.SystemMessage(system_prompt)] + messages

    ai_message = await model.ainvoke(messages)
    tool_messages = await langgraph.prebuilt.ToolNode(tools).ainvoke([ai_message])
    new_messages = [ai_message] + tool_messages
    return new_messages


class BotResponse(pydantic.BaseModel):
    last_user_message_language: str = pydantic.Field(
        description=inspect.cleandoc("""
            The language of the last user message, which the bot should also use for its own 
            response.
        """),
    )
    content: str = pydantic.Field(description="The message content that the user will see.")


def create_initial_agent(model, toolkit: JiraServiceDeskToolkit):
    """Create the agent for first invocation, after which the ReAct agent can be used."""

    async def entrypoint(
        state: langgraph.prebuilt.chat_agent_executor.AgentState,
    ):
        # TODO: can optimize by preemtively call RAG on raw user query,
        # and dropping it if AI decided to rerun it.
        system_prompt = inspect.cleandoc(f"""
            # Role
            You are a Jira service desk expert in a bank. Your goal is to help employees find the most 
            appropriate Jira request type for their problem.

            # Instructions
            Use the RAG retriever "{JiraServiceDeskToolkit.fetch_jira_request_type_guides_closest_to_user_query.__name__}" 
            tool to find relevant Jira request types for user's problem.
        """)  # noqa: E501
        new_messages = await ainvoke_model_with_tools(
            model,
            [toolkit.fetch_jira_request_type_guides_closest_to_user_query],
            state["messages"],
            system_prompt,
        )
        return {"messages": new_messages}

    async def thinking(state: langgraph.prebuilt.chat_agent_executor.AgentState):
        system_prompt = inspect.cleandoc(f"""
            # Role
            You are a Jira service desk expert in a bank. Your goal is to help employees find the most 
            appropriate Jira request type for their problem.

            Use the "{think.__name__}" tool to think and determine whether there is a request type 
            among the retrieved ones that satisfies the following requirements:
            - It clearly and unambiguously matches the user's problem.
            - Other request types are significantly less relevant.

            # Guidelines
            - Keep the reasoning concise and to the point, limit the thoughts to 100 words.
        """)  # noqa: E501
        new_messages = await ainvoke_model_with_tools(
            model,
            [think],
            state["messages"],
            system_prompt,
            tool_choice=think.__name__,
        )
        return {"messages": new_messages}

    async def route(state: langgraph.prebuilt.chat_agent_executor.AgentState):
        system_prompt = inspect.cleandoc(f"""
            # Role
            You are a Jira service desk expert in a bank. Your goal is to help employees find the most 
            appropriate Jira request type for their problem.

            # Choice
            - If there is no clear standout request type, then use the "{build_structured_response_tool(None).name}" tool
            to ask concise clarifying questions to gather the missing details and narrow down the request type. 
            Note that in this case you should detect the language of the user's very 
            last message (Azerbaijani, English, or Russian) and respond in that exact language.
            - Otherwise, call the "{fetch_detailed_jira_request_type_guide.__name__}" tool with the 
            best-matching request type.

            # Guidelines
            - Call exactly one tool out of those provided to you.
        """)  # noqa: E501
        tools = [
            fetch_detailed_jira_request_type_guide,
            build_structured_response_tool(BotResponse),
        ]
        new_messages = await ainvoke_model_with_tools(
            model, tools, state["messages"], system_prompt
        )
        return {"messages": new_messages}

    async def format_request_recommendation(
        state: langgraph.prebuilt.chat_agent_executor.AgentState,
    ):
        system_prompt = inspect.cleandoc(f"""
            # Role
            You are a Jira service desk expert in a bank. Your goal is to help employees find the most 
            appropriate Jira request type for their problem.

            # Language instruction
            - Your primary instruction, above all else, is to detect the language of the user's very 
            last message (Azerbaijani, English, or Russian) and ALWAYS respond in that exact language.
            - DO NOT deviate from this rule, even if other context suggests a different language might be helpful.

            # Output
            - Respond on the first line with a one-sentence recommendation that names the request type and why it fits.
            - On the second line, provide the non-prefilled request type URL from the tool output.
            - Then briefly provide and format three sections:
                - Self-help: how the user may try to resolve the issue independently.
                - Resolution pipeline: how this request is typically handled to resolution.
                - Required information: what is needed to create the request.
            - Proactively ask for user's feedback.
            - Use the "{build_structured_response_tool(None).name}" tool to generate the message for the user.

            # Examples
            ## Example 1
            User: '''
                I need to report a fraud related to my account.
            '''
            Bot: '''
                The 'Fraud' request type can be used to report suspected fraudulent activity affecting your account or transactions.
                https://jira-support.kapitalbank.az/servicedesk/customer/portal/1/create/1

                ### Self-help
                Immediately lock your card in the mobile app and review recent transactions.
                
                ### Resolution pipeline
                Triage → Transaction review by Fraud Ops → Temporary block if needed → Investigation → Customer notification and resolution.
                
                ### Required information for creating the request
                - Affected account/card
                - Transaction IDs and timestamps
                - Merchant name/location
            '''
        """)  # noqa: E501
        tools = [
            build_structured_response_tool(BotResponse),
        ]
        new_messages = await ainvoke_model_with_tools(
            model, tools, state["messages"], system_prompt
        )
        return {"messages": new_messages}

    def conditional_edge(state: langgraph.prebuilt.chat_agent_executor.AgentState):
        """If response tool was called, return END, otherwise go to format node."""
        last_message = state["messages"][-1]
        if not isinstance(last_message, langchain_core.messages.ToolMessage):
            raise ValueError("Last message is not a tool message.")

        if last_message.name == build_structured_response_tool(None).name:
            return langgraph.graph.END
        else:
            return format_request_recommendation.__name__

    graph = langgraph.graph.StateGraph(langgraph.prebuilt.chat_agent_executor.AgentStatePydantic)
    graph.add_node(entrypoint.__name__, entrypoint)
    graph.add_node(thinking.__name__, thinking)
    graph.add_node(route.__name__, route)
    graph.add_node(format_request_recommendation.__name__, format_request_recommendation)

    graph.set_entry_point(entrypoint.__name__)
    graph.add_edge(entrypoint.__name__, thinking.__name__)
    graph.add_edge(thinking.__name__, route.__name__)
    graph.add_conditional_edges(
        route.__name__,
        conditional_edge,
        # Specify the output nodes for visualization of the graph.
        [
            format_request_recommendation.__name__,
            langgraph.graph.END,
        ],
    )
    graph.set_finish_point(format_request_recommendation.__name__)

    compiled_graph = graph.compile()
    return compiled_graph



import contextlib

import gradio as gr
import numpy as np
import torch
import transformers

from .backend import Audio


def get_kokoro_sample_rate_hz():
    """https://github.com/PierrunoYT/Kokoro-TTS-Local/blob/master/README.md#:~:text=Sample%20rate:%2024kHz"""
    return 24_000


class TextToSpeechKokoro:
    """May require the espeak-ng CLI command to be installed."""

    def __init__(self, lang_code: str = "a", device: str = "cpu"):
        """Currently doesn't support mps device."""
        import kokoro  # type: ignore # Requires a lot of extra dependencies to run locally.

        self.pipeline = kokoro.KPipeline(lang_code=lang_code, device=device)

    def generate(
        self,
        text: str | list[str],
        voice: str = "af_heart",
        speed: int = 1,
        split_pattern: str = r"\n+",
    ):
        generator = self.pipeline(
            text,
            voice=voice,
            speed=speed,
            split_pattern=split_pattern,
        )
        for graphemes, phonemes, audio_tensor in generator:
            audio = audio_tensor.numpy()
            yield audio

    def to_audio_sample(self, text, **kwargs) -> Audio:
        audio = list(self.generate(text, **kwargs))
        audio = np.concatenate(audio)
        audio = gr.processing_utils.convert_to_16_bit_wav(audio)
        audio = Audio(rate=get_kokoro_sample_rate_hz(), data=audio)
        return audio


class SpeechToTextWhisper:
    """
    Currently local whisper is 2-3 times slower than OpenAI API,
    but it can be sped up using torch compile, quantization, different Python or C realizations.
    For example, using faster-whisper package (which may need the additional ffmpeg CLI dependency),
    which currently throws an error:

        audio: Audio
        byte_io = io.BytesIO()
        audio.write_wav_to_file(byte_io)
        blob = langchain_core.documents.base.Blob(data=byte_io.read(), mimetype="audio/wav")
        model = langchain_community.document_loaders.parsers.audio.FasterWhisperParser()
        output = model.parse(blob)
    """

    def __init__(
        self,
        device: str = "cpu",
        model_id: str = "openai/whisper-large-v3",
        use_torch_compile: bool = False,
    ):
        torch_dtype = torch.float32 if device == "cpu" else torch.float16
        model = transformers.AutoModelForSpeechSeq2Seq.from_pretrained(
            model_id,
            torch_dtype=torch_dtype,
            low_cpu_mem_usage=True,
            use_safetensors=True,
        )
        model.to(device)

        if use_torch_compile:
            torch.set_float32_matmul_precision("high")

            model.generation_config.cache_implementation = "static"
            model.generation_config.max_new_tokens = 256
            model.forward = torch.compile(model.forward, mode="reduce-overhead", fullgraph=True)

            torch._dynamo.config.suppress_errors = True

        processor = transformers.AutoProcessor.from_pretrained(model_id)
        pipeline = transformers.pipeline(
            "automatic-speech-recognition",
            model=model,
            tokenizer=processor.tokenizer,
            feature_extractor=processor.feature_extractor,
            torch_dtype=torch_dtype,
            device=device,
        )

        self.use_torch_compile = use_torch_compile
        self.model = model
        self.pipeline = pipeline

    def from_audio_sample(self, audio: Audio, language: str | None = None):
        generate_kwargs = {
            "return_timestamps": True,  # To enable processing of audio longer than 30 seconds
            "language": language,  # Specify input language for deterministic results
            # "task": "translate", # To translate input language to English
        }
        if self.use_torch_compile:
            generate_kwargs |= {
                "min_new_tokens": self.model.generation_config.max_new_tokens,
                "max_new_tokens": self.model.generation_config.max_new_tokens,
            }
            context = torch.nn.attention.sdpa_kernel(torch.nn.attention.SDPBackend.MATH)
        else:
            context = contextlib.nullcontext()

        with context:
            output = self.pipeline(
                audio.to_whisper_transformer_format(), generate_kwargs=generate_kwargs
            )

        text = output["text"]
        return text



import typer
import uvicorn

from ai_app.config import Stage, get_config


def launch_app(environment: Stage = Stage.local, workers: int = 1):
    uvicorn.run(
        "ai_app.app:build_app",  # String definition is required for reload and multiple workers.
        host="0.0.0.0" if environment == Stage.prod else "localhost",
        port=get_config().port,
        workers=workers,
        reload=environment == Stage.local,
        factory=True,
    )


def main():
    typer.run(launch_app)


if __name__ == "__main__":
    main()



import datetime
import itertools

import alembic.command
import polars as pl
import prefect

from ai_app.apps.jira_service_desk.preparation import (
    update_ai_jira_request_guides,
    update_jira_requests,
    update_manual_jira_requests,
)
from ai_app.config import (
    get_ai_postgres_engine,
    get_alembic_config,
    get_chat_model,
    get_config,
    get_engine,
    get_jira_service_desk_spreadsheets,
)
from ai_app.sql import setup_pgvector, setup_user


def initial_setup():
    admin_engine = get_engine("postgres_ai_admin")
    setup_pgvector(admin_engine)
    engine = get_ai_postgres_engine()
    setup_user(admin_engine, user=engine.url.username, password=engine.url.password)


def alembic_upgrade():
    alembic_config = get_alembic_config()
    alembic.command.upgrade(alembic_config, "head")


@prefect.task
def sync_dwh_tables(connection_schema_tables: dict[str, dict[str, list[str]]]):
    """Sync SQL tables from DWH to Postgres"""
    all_tables = list(
        itertools.chain(
            *(
                list(itertools.chain(*schema_tables.values()))
                for schema_tables in connection_schema_tables.values()
            )
        )
    )
    if len(all_tables) != len(set(all_tables)):
        raise ValueError("Table names are not unique")

    output_engine = get_engine("postgres_ai_dwh")
    for connection, schema_tables in connection_schema_tables.items():
        input_engine = get_engine(connection)
        for schema, tables in schema_tables.items():
            for table in tables:
                df = pl.read_database(
                    f"SELECT * from {schema}.{table}", input_engine, infer_schema_length=None
                )
                df.write_database(table, output_engine, if_table_exists="replace")


@prefect.flow
def update_jira_requests_meta_and_manual_data():
    get_config().setup_logging()
    prefect.task(update_jira_requests)()
    prefect.task(update_manual_jira_requests)(
        spreadsheets=get_jira_service_desk_spreadsheets(),
        write_back_to_gsheet=True,
    )


@prefect.flow
def update_ai_jira_request_summaries_and_embeddings():
    """
    Note that after adding new request types to AI Gsheet, they will have do_recommend=0,
    thus they will not be recommended to users until the field is manually set to 1.
    """
    get_config().setup_logging()
    prefect.task(update_ai_jira_request_guides)(
        model=get_chat_model("gpt-5-mini"),
        limit_requests=100,
    )


def main():
    prefect.serve(
        update_jira_requests_meta_and_manual_data.to_deployment(
            "update_jira_requests_meta_and_manual_data",
            interval=datetime.timedelta(hours=1),
            paused=True,
        ),
        update_ai_jira_request_summaries_and_embeddings.to_deployment(
            "update_ai_jira_request_summaries_and_embeddings",
            interval=datetime.timedelta(days=1),
            paused=True,
        ),
    )


if __name__ == "__main__":
    main()



import collections
import logging
from typing import Any

import langchain.tools
import langchain_core.messages
import langchain_core.tools
import pydantic

from ai_app.tools.atlassian import ConfluenceToolkit, JiraToolkit
from ai_app.tools.gitlab import GitLabToolkit
from ai_app.tools.meta import (
    InfoSecUserReminder,
    MetaToolkit,
    report_disclosure_of_sensitive_information,
    report_feedback,
    report_masked_entities_difficulties,
)
from ai_app.tools.sonarqube import SonarQubeToolkit
from ai_app.tools.sql import SqlToolkit


def extract_tool_artifacts_from_agent_response(
    messages: list[langchain_core.messages.BaseMessage],
) -> dict[str, list]:
    artifacts = collections.defaultdict(list)
    for message in messages:
        if isinstance(message, langchain_core.messages.ToolMessage) and message.artifact:
            artifacts[message.name].append(message.artifact)

    artifacts = dict(artifacts)
    return artifacts


def build_structured_response_tool(Response: type[pydantic.BaseModel]):
    """
    Structured output tool can decrease the number of model calls, but may be less reliable.
    It has additional benefit that the model sees the structured response schema in the ReAct loop,
    while the builtin implementation provides it to the agent after exiting the ReAct loop.
    Note that the docstring of the Response class is NOT passed to the model,
    only fields descriptions are.
    https://langchain-ai.github.io/langgraph/how-tos/react-agent-structured-output/#option-1-bind-output-as-tool
    """

    @langchain.tools.tool(
        args_schema=Response, return_direct=True, response_format="content_and_artifact"
    )
    def return_response(**kwargs) -> tuple[str, Response]:
        """Return response and end agent loop."""
        content = "Response was processed successfully."
        artifact = Response(**kwargs)
        return content, artifact

    return return_response


def get_structured_output_from_agent_response(response: dict) -> pydantic.BaseModel | None:
    if structured_response := response.get("structured_response"):
        return structured_response

    artifacts = extract_tool_artifacts_from_agent_response(response["messages"])
    artifacts = artifacts.get(build_structured_response_tool(None).name)
    # Take last structured response artifact.
    structured_response = artifacts[-1] if artifacts else None
    return structured_response


def think(thouhts: str) -> None:
    """This tool can be used to reason about the next steps and to create a plan."""
    # Note that since models generate tool calls independently and in parallel, this tool
    # call doesn't affect other calls. So if the thinking is required (for non-reasoning models)
    # before calling tools, then either the tools themselves should have "thoughts" as first
    # argument, or a separate reasoning API call should be made beforehand.



import functools
import inspect
import json
from typing import Annotated, override

import langchain.tools
import langchain_postgres
import logfire
import pydantic

from ai_app.ai_utils import json_dumps_for_ai
from ai_app.apps.jira_service_desk.data import (
    JiraRequest,
    JiraRequestCommonColumns,
    JiraRequestPrimaryKey,
    JiraRequestType,
    fetch_guides,
    fetch_use_case_guides_for_requests,
    get_index_column_names,
)
from ai_app.config import get_ai_postgres_rag_engine, get_config
from ai_app.rag import build_postgres_vector_store
from ai_app.tools.base import BaseToolkit


def build_postgres_jira_request_use_cases_ai_summary_vector_store(
    **kwargs,
) -> langchain_postgres.PGVectorStore:
    metadata_columns = [
        langchain_postgres.Column(c, data_type="int") for c in get_index_column_names()
    ]
    store = build_postgres_vector_store(
        engine=get_ai_postgres_rag_engine(),
        openai_api_key=get_config().secrets.openai_api_key,
        table_name="jira_request_use_cases_ai_summary",
        metadata_columns=metadata_columns,
        ensure_metadata_columns_index_is_unique=True,
        **kwargs,
    )
    return store


@functools.cache
def get_postgres_vector_store() -> langchain_postgres.PGVectorStore:
    return build_postgres_jira_request_use_cases_ai_summary_vector_store()


class JiraServiceDeskToolkit(BaseToolkit):
    def __init__(self, n_closest_requests: int = 5, service_desks: list[int] | None = None):
        super().__init__()
        self.n_closest_requests = n_closest_requests
        self.service_desks = service_desks

    @override
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        tools = [
            self.fetch_jira_request_type_guides_closest_to_user_query,
            fetch_detailed_jira_request_type_guide,
            build_url_for_jira_request_type,
        ]
        tools = [langchain.tools.tool(t) for t in tools]
        return tools

    @logfire.instrument()
    async def fetch_jira_request_type_guides_closest_to_user_query(
        self, query_in_english: str
    ) -> str:
        """Fetches Jira request type guides with use cases closest to user query via RAG."""
        store = get_postgres_vector_store()
        metadata_filter = (
            dict(service_desk_id={"$in": self.service_desks}) if self.service_desks else None
        )
        with logfire.span("Postgres vector store similarity search"):
            documents = await store.asimilarity_search(
                query_in_english, k=self.n_closest_requests, filter=metadata_filter
            )

        primary_keys = [JiraRequestPrimaryKey(**d.metadata) for d in documents]
        guides = fetch_use_case_guides_for_requests(primary_keys)
        guides = json_dumps_for_ai(guides)
        return guides


def fetch_detailed_jira_request_type_guide(service_desk_id: int, request_type_id: int) -> str:
    """
    Fetches Jira request type guide with information about when to use it and how the requests of
    this type are usually resolved. Additionally may return manually specified details which
    are more important than others, and in case of contradictions, the manually specified
    details take precedence.
    """
    primary_keys = [
        JiraRequestPrimaryKey(service_desk_id=service_desk_id, request_type_id=request_type_id)
    ]
    guides: list[JiraRequestType] = fetch_guides(primary_keys, with_input_fields=True)
    if len(guides) != 1:
        raise RuntimeError(f"Expected to fetch 1 guide, but got {len(guides)}")

    guide = guides[0]
    ai_input_fields = {field.field_id: field.description for field in guide.ai.input_request_fields}
    input_fields = [
        field.model_dump(exclude=list(JiraRequestCommonColumns.model_fields))
        | {"ai_generated_description": ai_input_fields.get(field.field_id)}
        for field in guide.meta.input_request_fields
    ]
    guide = dict(
        service_desk_id=guide.meta.service_desk_id,
        request_type_id=guide.meta.request_type_id,
        request_type_name=guide.meta.name,
        request_url=guide.meta.build_url(),
        use_case_summary=guide.ai.use_case,
        manually_specified_use_case_overrides=guide.manual.use_case,
        resolution_pipeline=guide.ai.resolution_pipeline,
        self_help=guide.ai.self_help,
        manually_specified_resolution_overrides=guide.manual.resolution,
        input_fields=input_fields,
    )
    guide = json_dumps_for_ai(guide)
    return guide


def build_url_for_jira_request_type(
    service_desk_id: int,
    request_type_id: int,
    request_type_name: str,
    # OpenAI models struggled to generate tool calls for `dict[str, str]` or `**kwargs: str`
    # arguments, so I switched to a more straightforward single JSON string argument instead.
    input_request_fields_json: Annotated[
        str,
        pydantic.Field(
            description=inspect.cleandoc(
                """
                The JSON with string key-value pairs for the Jira request input fields.
                The keys are the internal input_fields' field_ids taken from the output of the
                fetch_detailed_jira_request_type_guide tool, for which the
                can_be_passed_as_url_parameter is true.
                The values should align with the input field description and valid_values list,
                if present.
                """
            ),
        ),
    ],
) -> str:
    """
    Build a markdown URL for a Jira request type.
    The input field values should be provided in the input_request_fields_json argument.
    The URL will redirect to the Jira request creation page with prefilled values for these fields.
    """
    jira_request = JiraRequest(
        service_desk_id=service_desk_id,
        request_type_id=request_type_id,
        name=request_type_name,
    )
    input_request_fields = json.loads(input_request_fields_json)
    url = jira_request.build_markdown_url(**input_request_fields)
    return url



import inspect

import gradio as gr
import langchain_core.language_models
import langchain_core.messages
import langgraph.prebuilt

from ai_app.ai_utils import PreModelHook, astream_response_messages
from ai_app.config import get_confluence_toolkit, get_jira_toolkit
from ai_app.core import BaseApp, Response, State
from ai_app.frontend import build_model_choice_dropdown


def get_system_prompt():
    prompt = inspect.cleandoc("""
        # Role and Objective
        You are an IT Governance expert in a bank that can answer questions based on 
        corporate Jira and Confluence.

        # Guidelines
        - You are an agent - please keep going until the user's query is completely resolved, 
        before ending your turn and yielding back to the user.
        - Only terminate your turn when you are sure that the problem is solved.
        - Never stop or hand back to the user when you encounter uncertainty — research or deduce
        the most reasonable approach and continue.
        - Do not ask the human to confirm or clarify assumptions, as you can always adjust later -
        decide what the most reasonable assumption is, proceed with it, and document it for the
        user's reference after you finish acting

        # Output Format
        Use Markdown to format your response.
    """)
    return prompt


class App(BaseApp):
    name = "Governance agent"
    requires_auth = False

    async def astream(
        self,
        model: langchain_core.language_models.BaseChatModel,
        message: langchain_core.messages.BaseMessage,
        frontend_messages: list[langchain_core.messages.BaseMessage],
        state: State,
    ):
        messages = state.messages + [message]
        agent = langgraph.prebuilt.create_react_agent(
            model=model,
            tools=get_jira_toolkit().get_tools() + get_confluence_toolkit().get_tools(),
            pre_model_hook=PreModelHook(model),
            prompt=get_system_prompt(),
        )
        async for response in astream_response_messages(agent, messages, stream_ai_tokens=True):
            yield Response(
                content=response.formatted_response,
                messages=response.state["messages"],
            )

    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks(fill_height=True) as app:
            model_choice = model_choice or build_model_choice_dropdown()
            self.build_gradio_chat_interface(
                model_choice=model_choice,
                description=("The chatbot can answer questions based on Jira and Confluence"),
                examples=[
                    [
                        "What is the current focus of the ACE Jira project?",
                    ],
                    [
                        "What is Jira project ACE about?",
                    ],
                    [
                        "Which tools do you have access to?",
                    ],
                ],
            )

        return app
